# Licensed under an MIT open source license - see LICENSE

import numpy as np
import astropy.units as u
import networkx as nx
import warnings
import scipy.ndimage as nd
from astropy.nddata import extract_array
import astropy.modeling as mod
from astropy.modeling.models import Gaussian1D, Const1D
import sys

if sys.version_info[0] >= 3:
    import _pickle as pickle
else:
    import cPickle as pickle

from .length import (init_lengths, main_length, make_final_skeletons,
                     pre_graph, longest_path, prune_graph)
from .pixel_ident import pix_identify
from .utilities import pad_image, in_ipynb, red_chisq
from .base_conversions import UnitConverter
from .rollinghough import rht
from .width import (radial_profile, gaussian_model, fit_radial_model,
                    nonparam_width)


class FilamentNDBase(object):
    """
    Analysis and properties of a single filament object.
    """
    @property
    def pixel_coords(self):
        return self._pixel_coords

    @property
    def pixel_extents(self):
        return [tuple([coord.min() for coord in self._orig_pixel_coords]),
                tuple([coord.max() for coord in self._orig_pixel_coords])]

    def position(self, world_coord=False):
        '''
        Return the centre position of the filament based on the pixel
        coordinates.
        '''
        centres = [np.median(coord) for coord in self._orig_pixel_coords]

        if world_coord:
            if hasattr(self._converter, '_wcs'):
                wcs = self._converter._wcs
                # Convert to world coordinates
                posn_tuple = centres + [0]
                w_centres = wcs.all_pix2world(*posn_tuple)

                # Attach units
                wu_centres = [val * u.Unit(wcs.wcs.cunit[i]) for i, val
                              in enumerate(w_centres)]
                return wu_centres
            else:
                warnings.warn("No WCS information given. Returning pixel"
                              " position.")
                return [centre * u.pix for centre in centres]
        else:
            return [centre * u.pix for centre in centres]


class Filament2D(FilamentNDBase):
    """
    Analysis and properties of a 2D filament.

    Parameters
    ----------
    pixel_coords : tuple of `~np.ndarray`
        Pixel coordinates as a set of arrays (i.e., the output from
        `~numpy.where`).
    converter : `~fil_finder.base_conversions.UnitConverter`, optional
        Unit converter class.
    wcs : `~astropy.wcs.WCS`, optional
        WCS information for the pixel set.
    distance : `~astropy.units.Quantity`, optional
        Distance to the region described by the pixel set. Requires for
        conversions to physical units.
    """
    def __init__(self, pixel_coords, converter=None, wcs=None, distance=None):
        super(Filament2D, self).__init__()

        self._pixel_coords = pixel_coords

        # Create a separate account of the initial skeleton pixels
        self._orig_pixel_coords = pixel_coords

        if converter is not None:
            self._converter = converter
        else:
            self._converter = UnitConverter(wcs=wcs, distance=distance)

    def image_slicer(self, image, out_shape, pad_size=0):
        '''
        Create a cut-out of a given image to some output shape with optional
        padding on the edges. The given image must be on the same pixel grid
        as the image used to create the skeleton.

        Parameters
        ----------
        image : `~numpy.ndarray` or `~astropy.units.Quantity`
            Image to slice out around the skeleton.
        out_shape : tuple
            2D output shape.
        pad_size : int, optional
            Number of pixels to pad.

        Returns
        -------
        out_arr : `~numpy.ndarray` or `~astropy.units.Quantity`
            Output array with given shape.
        '''

        arr_cent = [(out_shape[0] - pad_size * 2 - 1) / 2. +
                    self.pixel_extents[0][0],
                    (out_shape[1] - pad_size * 2 - 1) / 2. +
                    self.pixel_extents[0][1]]

        out_arr = extract_array(image, out_shape, arr_cent)

        # astropy v4.0 now retains the unit. So only add a unit
        # when out_arr isn't a Quantity
        if hasattr(image, "unit") and not hasattr(out_arr, 'unit'):
            out_arr = out_arr * image.unit

        return out_arr

    def skeleton(self, pad_size=0, corner_pix=None, out_type='all'):
        '''
        Create a mask from the pixel coordinates.

        Parameters
        ----------
        pad_size : int, optional
            Number of pixels to pad along each edge.
        corner_pix : tuple of ints, optional
            The position of the left-bottom corner of the pixels in the
            skeleton. Used for offsetting the location of the pixels.
        out_type : {"all", "longpath"}, optional
            Return the entire skeleton or just the longest path. Default is to
            return the whole skeleton.

        Returns
        -------
        mask : `~numpy.ndarray`
            Boolean mask containing the skeleton pixels.
        '''

        pad_size = int(pad_size)
        if pad_size < 0:
            raise ValueError("pad_size must be a positive integer.")

        if corner_pix is None:
            # Place the smallest pixel in the set at the pad size
            corner_pix = [pad_size, pad_size]

        out_types = ['all', 'longpath']
        if out_type not in out_types:
            raise ValueError("out_type must be 'all' or 'longpath'.")

        y_shape = self.pixel_extents[1][0] - self.pixel_extents[0][0] + \
            2 * pad_size + 1
        x_shape = self.pixel_extents[1][1] - self.pixel_extents[0][1] + \
            2 * pad_size + 1

        mask = np.zeros((y_shape, x_shape), dtype=bool)

        if out_type == 'all':
            pixels = self.pixel_coords
        else:
            if not hasattr(self, '_longpath_pixel_coords'):
                raise AttributeError("longest path is not defined. Run "
                                     "`Filament2D.skeleton_analysis` first.")
            pixels = self.longpath_pixel_coords

        mask[pixels[0] - self.pixel_extents[0][0] + corner_pix[0],
             pixels[1] - self.pixel_extents[0][1] + corner_pix[1]] = True

        return mask

    def skeleton_analysis(self, image, verbose=False, save_png=False,
                          save_name=None, prune_criteria='all',
                          relintens_thresh=0.2, max_prune_iter=10,
                          branch_thresh=0 * u.pix):
        '''
        Run the skeleton analysis.

        Separates skeleton structures into branches and intersections. Branches
        below the pruning criteria are removed. The structure is converted into
        a graph object to find the longest path. The pruned skeleton is used in
        the subsequent analysis steps.

        Parameters
        ----------
        image : `~numpy.ndarray` or `~astropy.units.Quantity`
            Data the filament was extracted from.
        verbose : bool, optional
            Show intermediate plots.
        save_png : bool, optional
            Save the plots in verbose mode.
        save_name : str, optional
            Prefix for the saved plots.
        prune_criteria : {'all', 'intensity', 'length'}, optional
            Choose the property to base pruning on. 'all' requires that the
            branch fails to satisfy the length and relative intensity checks.
        relintens_thresh : float, optional
            Value between 0 and 1 that sets the relative importance of the
            intensity-to-length criteria when pruning. Only used if
            `prune_criteria='all'`.
        max_prune_iter : int, optional
            Maximum number of pruning iterations to apply.
        branch_thresh : `~astropy.units.Quantity`, optional
            Minimum length for a branch to be eligible to be pruned.
        '''

        # NOTE:
        # All of these functions are essentially the same as those used for
        # fil_finder_2D. For now, they all are expecting lists with each
        # filament property as an element. Everything is wrapped to be a list
        # because of this, but will be removed once fil_finder_2D is removed.
        # A lot of this can be streamlined in that process.

        if save_png and save_name is None:
            raise ValueError("save_name must be given when save_png=True.")

        # Must have a pad size of 1 for the morphological operations.
        pad_size = 1
        self._pad_size = pad_size

        branch_thresh = self._converter.to_pixel(branch_thresh)

        # Do we need to pad the image before slicing?
        input_image = pad_image(image, self.pixel_extents, pad_size)

        skel_mask = self.skeleton(pad_size=pad_size)

        # If the padded image matches the mask size, don't need additional
        # slicing
        if input_image.shape != skel_mask.shape:
            input_image = self.image_slicer(input_image, skel_mask.shape,
                                            pad_size=pad_size)

        # The mask and sliced image better have the same shape!
        if input_image.shape != skel_mask.shape:
            raise AssertionError("Sliced image shape does not equal the mask "
                                 "shape. This should never happen! If you see"
                                 " this issue, please report it as a bug!")

        iter = 0

        while True:

            skel_mask = self.skeleton(pad_size=pad_size)

            interpts, hubs, ends, filbranches, labeled_mask =  \
                pix_identify([skel_mask], 1)

            branch_properties = init_lengths(labeled_mask, filbranches,
                                             [[(0, 0), (0, 0)]],
                                             input_image)

            edge_list, nodes, loop_edges = \
                pre_graph(labeled_mask, branch_properties, interpts, ends)

            max_path, extremum, G = \
                longest_path(edge_list, nodes,
                             verbose=False,
                             skeleton_arrays=labeled_mask)

            # Skip pruning if skeleton has only one branch
            if len(G[0].nodes()) > 1:
                updated_lists = \
                    prune_graph(G, nodes, edge_list, max_path, labeled_mask,
                                branch_properties, loop_edges,
                                prune_criteria=prune_criteria,
                                length_thresh=branch_thresh.value,
                                relintens_thresh=relintens_thresh,
                                max_iter=1)
                labeled_mask, edge_list, nodes, branch_properties = \
                    updated_lists

            final_fil_arrays =\
                make_final_skeletons(labeled_mask, interpts,
                                     verbose=False)

            # Update the skeleton pixels
            good_pix = np.where(final_fil_arrays[0])
            self._pixel_coords = \
                (good_pix[0] + self.pixel_extents[0][0] - pad_size,
                 good_pix[1] + self.pixel_extents[0][1] - pad_size)

            if iter == 0:
                prev_G = G[0]
                iter += 1
                if iter == max_prune_iter:
                    break
                else:
                    continue

            # Isomorphic comparison is failing for networkx 2.1
            # I don't understand the error, so we'll instead require
            # that the nodes be the same. This should be safe as
            # pruning can only remove nodes.

            # edge_match = iso.numerical_edge_match('weight', 1)
            # if nx.is_isomorphic(prev_G, G[0],
            #                     edge_match=edge_match):

            # the node attribute was removed in 2.4.
            if hasattr(G, 'node'):
                if prev_G.node == G[0].node:
                    break

            if hasattr(G, 'nodes'):
                if prev_G.nodes == G[0].nodes:
                    break

            prev_G = G[0]
            iter += 1

            if iter >= max_prune_iter:
                warnings.warn("Graph pruning reached max iterations.")
                break

        self._graph = G[0]

        # Run final analyses for plotting, etc.
        max_path, extremum, G = \
            longest_path(edge_list, nodes,
                         verbose=verbose,
                         save_png=save_png,
                         save_name="{0}_graphstruct.png".format(save_name),
                         skeleton_arrays=labeled_mask)

        length_output = main_length(max_path, edge_list, labeled_mask,
                                    interpts,
                                    branch_properties["length"],
                                    1.,
                                    verbose=verbose, save_png=save_png,
                                    save_name="{0}_longestpath.png".format(save_name))
        lengths, long_path_array = length_output

        good_long_pix = np.where(long_path_array[0])
        self._longpath_pixel_coords = \
            (good_long_pix[0] + self.pixel_extents[0][0] - pad_size,
             good_long_pix[1] + self.pixel_extents[0][1] - pad_size)

        self._length = lengths[0] * u.pix

        final_fil_arrays =\
            make_final_skeletons(labeled_mask, interpts,
                                 verbose=verbose, save_png=save_png,
                                 save_name="{0}_finalskeleton.png".format(save_name))

        # Track the final intersection and end points
        interpts, hubs, ends =  \
            pix_identify([final_fil_arrays[0].copy()], 1)[:3]

        # Adjust intersection and end points to be in the original array
        # positions
        corr_inters = []
        for inter in interpts[0]:
            per_inter = []

            for ints in inter:
                per_inter.append((ints[0] + self.pixel_extents[0][0] - pad_size,
                                  ints[1] + self.pixel_extents[0][1] - pad_size))

            corr_inters.append(per_inter)
        self._interpts = corr_inters

        corr_ends = []
        for end in ends[0]:
            corr_ends.append((end[0] + self.pixel_extents[0][0] - pad_size,
                              end[1] + self.pixel_extents[0][1] - pad_size))
        self._endpts = corr_ends

        # Update the skeleton pixels
        good_pix = np.where(final_fil_arrays[0])
        self._pixel_coords = \
            (good_pix[0] + self.pixel_extents[0][0] - pad_size,
             good_pix[1] + self.pixel_extents[0][1] - pad_size)

        self._branch_properties = \
            {'length': branch_properties['length'][0] * u.pix,
             'intensity': np.array(branch_properties['intensity'][0]),
             'number': branch_properties['number'][0],
             'pixels': branch_properties['pixels'][0]}

    @property
    def branch_properties(self):
        '''
        Dictionary with branch lengths, average intensity, and pixels.
        '''
        return self._branch_properties

    def branch_pts(self, img_coords=False):
        '''
        Pixels within each skeleton branch.

        Parameters
        ----------
        img_coords : bool
            Return the branch pts in coordinates of the original image.
        '''
        if not img_coords:
            return self.branch_properties['pixels']

        # Transform from per-filament to image coords
        img_branch_pts = []
        for bpts in self.branch_properties['pixels']:

            bpts_copy = bpts.copy()

            bpts_copy[:, 0] = bpts[:, 0] + self.pixel_extents[0][0] - self._pad_size
            bpts_copy[:, 1] = bpts[:, 1] + self.pixel_extents[0][1] - self._pad_size

            img_branch_pts.append(bpts_copy)

        return img_branch_pts


    @property
    def intersec_pts(self):
        '''
        Skeleton pixels associated intersections.
        '''
        return self._interpts

    @property
    def end_pts(self):
        '''
        Skeleton pixels associated branch end.
        '''
        return self._endpts

    def length(self, unit=u.pixel):
        '''
        The longest path length of the skeleton

        Parameters
        ----------
        unit : `~astropy.units.Unit`, optional
            Pixel, angular, or physical unit to convert to.
        '''
        return self._converter.from_pixel(self._length, unit)

    @property
    def longpath_pixel_coords(self):
        '''
        Pixel coordinates of the longest path.
        '''
        return self._longpath_pixel_coords

    @property
    def graph(self):
        '''
        The networkx graph for the filament.
        '''
        return self._graph

    def plot_graph(self, save_name=None, layout_func=nx.spring_layout):
        '''
        Plot the graph structure.

        Parameters
        ----------
        save_name : str, optional
            Name of saved plot. A plot is only saved if a name is given.
        layout_func : networkx layout function, optional
            Layout function from networkx. Defaults to `spring_layout`.
        '''
        import matplotlib.pyplot as plt

        G = self.graph

        elist = [(u, v) for (u, v, d) in G.edges(data=True)]
        posns = layout_func(G)

        nx.draw_networkx_nodes(G, posns, node_size=200)
        nx.draw_networkx_edges(G, posns, edgelist=elist, width=2)
        nx.draw_networkx_labels(G, posns, font_size=10,
                                font_family='sans-serif')
        plt.axis('off')

        if save_name is not None:
            # Save the plot
            plt.savefig(save_name)
            plt.close()
        else:
            plt.show()

        # Add in the ipynb checker

    def rht_analysis(self, radius=10 * u.pix, ntheta=180,
                     background_percentile=25):
        '''
        Use the RHT to find the filament orientation and dispersion of the
        longest path.

        Parameters
        ----------
        radius : `~astropy.units.Quantity`, optional
            Radius of the region to compute the orientation within. Converted
            to pixel units and rounded to the nearest integer.
        ntheta : int, optional
            Number of angles to sample at. Default is 180.
        background_percentile : float, optional
            Float between 0 and 100 that sets a background level for the RHT
            distribution before calculating orientation and curvature.
        '''

        if not hasattr(radius, 'unit'):
            warnings.warn("Radius has no given units. Assuming pixel units.")
            radius *= u.pix

        radius = int(round(self._converter.to_pixel(radius).value))

        longpath_arr = self.skeleton(out_type='longpath')

        longpath_arr = np.fliplr(longpath_arr)

        theta, R, quant = rht(longpath_arr, radius, ntheta,
                              background_percentile)

        twofive, mean, sevenfive = quant

        self._orientation = mean * u.rad

        if sevenfive > twofive:
            self._curvature = np.abs(sevenfive - twofive) * u.rad
        else:
            self._curvature = (np.abs(sevenfive - twofive) + np.pi) * u.rad

        self._orientation_hist = [theta, R]
        self._orientation_quantiles = [twofive, sevenfive]

    @property
    def orientation_hist(self):
        '''
        Distribution of orientations from the RHT along the longest path.

        Contains the angles of the distribution bins and the values in those
        bins.
        '''
        return self._orientation_hist

    @property
    def orientation(self):
        '''
        Mean orientation of the filament along the longest path.
        '''
        return self._orientation

    @property
    def curvature(self):
        '''
        Interquartile range of the RHT orientation distribution along the
        longest path.
        '''
        return self._curvature

    def plot_rht_distrib(self, save_name=None):
        '''
        Plot the RHT distribution from `Filament2D.rht_analysis`.

        Parameters
        ----------
        save_name : str, optional
            Name of saved plot. A plot is only saved if a name is given.
        '''

        theta = self.orientation_hist[0]
        R = self.orientation_hist[1]

        import matplotlib.pyplot as plt

        median = self.orientation.value
        twofive, sevenfive = self._orientation_quantiles

        ax1 = plt.subplot(121, polar=True)
        ax1.plot(2 * theta, R / R.max(), "kD")
        ax1.fill_between(2 * theta, 0,
                         R[:, 0] / R.max(),
                         facecolor="blue",
                         interpolate=True, alpha=0.5)
        ax1.set_rmax(1.0)
        ax1.plot([2 * median] * 2, np.linspace(0.0, 1.0, 2), "g")
        ax1.plot([2 * twofive] * 2, np.linspace(0.0, 1.0, 2),
                 "b--")
        ax1.plot([2 * sevenfive] * 2, np.linspace(0.0, 1.0, 2),
                 "b--")
        plt.subplot(122)
        plt.imshow(self.skeleton(out_type='longpath'),
                   cmap="binary", origin="lower")
        if save_name is not None:
            plt.savefig(save_name)
            plt.close()
        else:
            plt.show()

    def rht_branch_analysis(self, radius=10 * u.pix, ntheta=180,
                            background_percentile=25,
                            min_branch_length=3 * u.pix):
        '''
        Use the RHT to find the filament orientation and dispersion of each
        branch in the filament.

        Parameters
        ----------
        radius : `~astropy.units.Quantity`, optional
            Radius of the region to compute the orientation within. Converted
            to pixel units and rounded to the nearest integer.
        ntheta : int, optional
            Number of angles to sample at. Default is 180.
        background_percentile : float, optional
            Float between 0 and 100 that sets a background level for the RHT
            distribution before calculating orientation and curvature.
        min_branch_length : `~astropy.units.Quantity`, optional
            Minimum length of a branch to run the RHT on. Branches that are
            too short will cause spikes along the axis angles or 45 deg. off.
        '''

        # Convert length cut to pixel units

        if not hasattr(radius, 'unit'):
            warnings.warn("Radius has no given units. Assuming pixel units.")
            radius *= u.pix

        if not hasattr(min_branch_length, 'unit'):
            warnings.warn("min_branch_length has no given units. Assuming "
                          "pixel units.")
            min_branch_length *= u.pix

        radius = int(round(self._converter.to_pixel(radius).value))
        min_branch_length = self._converter.to_pixel(min_branch_length).value

        means = []
        iqrs = []

        # Make padded arrays from individual branches
        for i, (pix, length) in enumerate(zip(self.branch_pts(img_coords=False),
                                              self.branch_properties['length'])):

            if length.value < min_branch_length:
                means.append(np.NaN)
                iqrs.append(np.NaN)
                continue

            # Setup size of array
            ymax = pix[:, 0].max()
            ymin = pix[:, 0].min()
            xmax = pix[:, 1].max()
            xmin = pix[:, 1].min()

            shape = (ymax - ymin + 1 + 2 * radius,
                     xmax - xmin + 1 + 2 * radius)

            branch_array = np.zeros(shape, dtype=bool)
            branch_array[pix[:, 0] - ymin + radius,
                         pix[:, 1] - xmin + radius] = True

            branch_array = np.fliplr(branch_array)

            theta, R, quant = rht(branch_array, radius, ntheta,
                                  background_percentile)

            twofive, mean, sevenfive = quant

            means.append(mean)
            if sevenfive > twofive:
                iqrs.append(np.abs(sevenfive - twofive))
            else:
                iqrs.append(np.abs(sevenfive - twofive) + np.pi)

        self._orientation_branches = np.array(means) * u.rad
        self._curvature_branches = np.array(iqrs) * u.rad

    @property
    def orientation_branches(self):
        '''
        Orientations along each branch in the filament.
        '''
        return self._orientation_branches

    @property
    def curvature_branches(self):
        '''
        Curvature along each branch in the filament.
        '''
        return self._curvature_branches

    def width_analysis(self, image, all_skeleton_array=None,
                       max_dist=10 * u.pix,
                       pad_to_distance=0 * u.pix,
                       fit_model='gaussian_bkg',
                       fitter=None,
                       try_nonparam=True,
                       use_longest_path=False,
                       add_width_to_length=False,
                       deconvolve_width=True,
                       beamwidth=None,
                       fwhm_function=None,
                       chisq_max=10.,
                       **kwargs):
        '''

        Create an average radial profile for the filament and fit a given
        model.

        Parameters
        ----------
        image : `~astropy.unit.Quantity` or `~numpy.ndarray`
            The image from which the filament was extracted.
        all_skeleton_array : np.ndarray
            An array with the skeletons of other filaments. This is used to
            avoid double-counting pixels in the radial profiles in nearby
            filaments.
        max_dist : `~astropy.units.Quantity`, optional
            Largest radius around the skeleton to create the profile from. This
            can be given in physical, angular, or physical units.
        pad_to_distance : `~astropy.units.Quantity`, optional
            Force all pixels within this distance to be kept, even if a pixel
            is closer to another skeleton, as given in `all_skeleton_array`.
        fit_model : str or `~astropy.modeling.Fittable1DModel`, optional
            The model to fit to the profile. Built-in models include
            'gaussian_bkg' for a Gaussian with a constant background,
            'gaussian_nobkg' for just a Gaussian, 'nonparam' for the
            non-parametric estimator. Defaults to 'gaussian_bkg'.
        fitter : `~astropy.modeling.fitting.Fitter`, optional
            One of the astropy fitting classes. Defaults to a
            Levenberg-Marquardt fitter.
        try_nonparam : bool, optional
            If the chosen model fit fails, fall back to a non-parametric
            estimate.
        use_longest_path : bool, optional
            Only fit profile to the longest path skeleton. Disabled by
            default.
        add_width_to_length : bool, optional
            Add the FWHM to the filament length. This accounts for the
            expected shortening in the medial axis transform. Enabled by
            default.
        deconvolve_width : bool, optional
            Deconvolve the beam width from the FWHM. Enabled by default.
        beamwidth : `~astropy.units.Quantity`, optional
            The beam width to deconvolve the FWHM from. Required if
            `deconvolve_width = True`.
        fwhm_function : function, optional
            Convert the width parameter to the FWHM. Must take the fit model
            as an argument and return the FWHM and its uncertainty. If no
            function is given, the Gaussian FWHM is used.
        chisq_max : float, optional
            Enable the fail flag if the reduced chi-squared value is above
            this limit.
        kwargs : Passed to `~fil_finder.width.radial_profile`.

        '''

        # Convert quantities to pixel units.
        max_dist = self._converter.to_pixel(max_dist).value
        pad_to_distance = self._converter.to_pixel(pad_to_distance).value

        if deconvolve_width and beamwidth is None:
            raise ValueError("beamwidth must be given when deconvolve_width is"
                             " enabled.")

        if beamwidth is not None:
            beamwidth = self._converter.to_pixel(beamwidth)

        # Use the max dist as the pad size
        pad_size = int(np.ceil(max_dist))

        # if given a master skeleton array, require it to be the same shape as
        # the image
        if all_skeleton_array is not None:
            if all_skeleton_array.shape != image.shape:
                raise ValueError("The shape of all_skeleton_array must match"
                                 " the given image.")

        if use_longest_path:
            skel_array = self.skeleton(pad_size=pad_size, out_type='longpath')
        else:
            skel_array = self.skeleton(pad_size=pad_size, out_type='all')

        out_shape = skel_array.shape
        input_image = self.image_slicer(image, out_shape, pad_size=pad_size)

        if all_skeleton_array is not None:
            input_all_skeleton_array = \
                self.image_slicer(all_skeleton_array, out_shape,
                                  pad_size=pad_size)
        else:
            input_all_skeleton_array = None

        # Create distance arrays to build profile from
        dist_skel_arr = nd.distance_transform_edt(np.logical_not(skel_array))

        # And create a distance array from the full skeleton array if given
        if input_all_skeleton_array is not None:
            dist_skel_all = nd.distance_transform_edt(np.logical_not(input_all_skeleton_array))
        else:
            dist_skel_all = None

        # Need the unbinned data for the non-parametric fit.
        out = radial_profile(input_image, dist_skel_all,
                             dist_skel_arr,
                             [(0, 0), (0, 0)],
                             max_distance=max_dist,
                             pad_to_distance=pad_to_distance,
                             **kwargs)

        if out is None:
            raise ValueError("Building radial profile failed. Check the input"
                             " image for NaNs.")
        else:
            dist, radprof, weights, unbin_dist, unbin_radprof = out

        # Attach units
        xunit = u.pix
        if hasattr(image, 'unit'):
            yunit = image.unit
        else:
            yunit = u.dimensionless_unscaled

        self._yunit = yunit

        radprof = radprof * yunit
        dist = dist * xunit

        self._radprofile = [dist, radprof]
        self._unbin_radprofile = [unbin_dist * xunit,
                                  unbin_radprof * yunit]

        # Make sure the given model is valid
        if not isinstance(fit_model, mod.Model):
            skip_fitting = False
            self._radprof_type = fit_model
            # Check the default types
            if fit_model == "gaussian_bkg":
                fit_model = gaussian_model(dist, radprof, with_bkg=True)
            elif fit_model == "gaussian_nobkg":
                fit_model = gaussian_model(dist, radprof, with_bkg=False)
            elif fit_model == "nonparam":
                skip_fitting = True
            else:
                raise ValueError("fit_model must be an "
                                 "astropy.modeling.Fittable1DModel or "
                                 "one of the default models: 'gaussian_bkg',"
                                 " 'gaussian_nobkg', or 'nonparam'.")
        else:
            # Record the fit type
            self._radprof_type = fit_model.name

        if not skip_fitting:
            fitted_model, fitter = fit_radial_model(dist, radprof, fit_model,
                                                    weights=weights)

            # Only keep the non-fixed parameters. The fixed parameters won't
            # appear in the covariance matrix.
            params = []
            names = []
            for name in fitted_model.param_names:
                # Check if it is fixed:
                if fitted_model.fixed[name]:
                    continue

                param = getattr(fitted_model, name)

                if param.quantity is not None:
                    params.append(param.quantity)
                else:
                    # Assign a dimensionless unit
                    params.append(param.value * u.dimensionless_unscaled)

                names.append(name)

            self._radprof_params = params
            npar = len(self.radprof_params)

            self._radprof_parnames = names

            self._radprof_model = fitted_model
            self._radprof_fitter = fitter

            # Fail checks
            fail_flag = False

            param_cov = fitter.fit_info.get('param_cov')
            if param_cov is not None:
                fit_uncert = list(np.sqrt(np.diag(param_cov)))
            else:
                fit_uncert = [np.NaN] * npar
                fail_flag = True

            if len(fit_uncert) != len(params):
                raise ValueError("The number of parameters does not match the "
                                 "number from the covariance matrix. Check for"
                                 " fixed parameters.")

            # Add units to errors
            for i, par in enumerate(params):
                fit_uncert[i] = fit_uncert[i] * par.unit

            self._radprof_errors = fit_uncert

            # Check if units should be kept
            if fitted_model._supports_unit_fitting:
                modvals = fitted_model(dist)
                radprof_vals = radprof
            else:
                modvals = fitted_model(dist.value)
                radprof_vals = radprof.value

            chisq = red_chisq(radprof_vals, modvals, npar, 1)
            if chisq > chisq_max:
                fail_flag = True

        if (skip_fitting or fail_flag) and try_nonparam:
            fit, fit_error, fail_flag = \
                nonparam_width(dist.value, radprof.value,
                               unbin_dist, unbin_radprof,
                               None, 5, 99)
            self._radprof_type = 'nonparam'

            # Make the equivalent Gaussian model w/ a background
            self._radprof_model = Gaussian1D() + Const1D()
            if self._radprof_model._supports_unit_fitting:
                self._radprof_model.amplitude_0 = fit[0] * yunit
                self._radprof_model.mean_0 = 0.0 * xunit
                self._radprof_model.sigma_0 = fit[1] * xunit
                self._radprof_model.amplitude_1 = fit[2] * yunit
            else:
                self._radprof_model.amplitude_0 = fit[0]
                self._radprof_model.mean_0 = 0.0
                self._radprof_model.sigma_0 = fit[1]
                self._radprof_model.amplitude_1 = fit[2]

            # Slice out the FWHM and add units
            params = [fit[0] * yunit, fit[1] * xunit, fit[2] * yunit]
            errs = [fit_error[0] * yunit, fit_error[1] * xunit,
                    fit_error[2] * yunit]
            self._radprof_params = params
            self._radprof_errors = errs
            self._radprof_parnames = ['amplitude_0', 'stddev_0', 'amplitude_1']

        if fwhm_function is not None:
            fwhm = fwhm_function(fitted_model)
        else:
            # Default to Gaussian FWHM
            for idx, name in enumerate(self.radprof_parnames):
                if "stddev" in name:
                    found_width = True
                    break

            if found_width:
                fwhm = self.radprof_params[idx].value * np.sqrt(8 * np.log(2)) * xunit
                fwhm_err = self.radprof_errors[idx].value * np.sqrt(8 * np.log(2)) * xunit
            else:
                raise ValueError("Could not automatically identify which "
                                 "parameter in the model corresponds to the "
                                 "width. Please pass a function to "
                                 "'fwhm_function' to identify the width "
                                 "parameter.")

        if deconvolve_width:
            fwhm_deconv_sq = fwhm**2 - beamwidth**2
            if fwhm_deconv_sq > 0:
                fwhm_deconv = np.sqrt(fwhm_deconv_sq)
                fwhm_deconv_err = fwhm * fwhm_err / fwhm_deconv
            else:
                fwhm_deconv = np.NaN
                fwhm_deconv_err = np.NaN
                warnings.warn("Width could not be deconvolved from the beam "
                              "width.")
        else:
            fwhm_deconv = fwhm
            fwhm_deconv_err = fwhm_err

        self._fwhm = fwhm_deconv
        self._fwhm_err = fwhm_deconv_err

        # Final width check -- make sure length is longer than the width.
        # If it is, add the width onto the length since the adaptive
        # thresholding shortens each edge by the about the same.
        if self.length() < self._fwhm:
            fail_flag = True

        # Add the width onto the length if enabled
        if add_width_to_length:
            if fail_flag:
                warnings.warn("Ignoring adding the width to the length because"
                              " the fail flag was raised for the fit.")
            else:
                self._length += self._fwhm

        self._radprof_failflag = fail_flag

    @property
    def radprof_fit_fail_flag(self):
        '''
        Flag to catch poor fits.
        '''
        return self._radprof_failflag

    @property
    def radprof_type(self):
        '''
        The model type used to fit the radial profile.
        '''
        return self._radprof_type

    @property
    def radprofile(self):
        '''
        The binned radial profile created in `~FilFinder2D.width_analysis`.
        This contains the distances and the profile value in the distance bin.
        '''
        return self._radprofile

    @property
    def radprof_params(self):
        '''
        Fit parameters from `~FilFinder2D.width_analysis`.
        '''
        return self._radprof_params

    @property
    def radprof_errors(self):
        '''
        Fit uncertainties from `~FilFinder2D.width_analysis`.
        '''
        return self._radprof_errors

    def radprof_fwhm(self, unit=u.pixel):
        '''
        The FWHM of the fitted radial profile and its uncertainty.

        Parameters
        ----------
        unit : `~astropy.units.Unit`, optional
            Pixel, angular, or physical unit to convert to.
        '''
        return self._converter.from_pixel(self._fwhm, unit), \
            self._converter.from_pixel(self._fwhm_err, unit)

    @property
    def radprof_parnames(self):
        '''
        Parameter names from `~FilFinder2D.radprof_model`.
        '''
        return self._radprof_parnames

    def radprof_fit_table(self, unit=u.pix):
        '''
        Return an `~astropy.table.Table` with the fit parameters and
        uncertainties.

        Parameters
        ----------
        unit : `~astropy.units.Unit`, optional
            Pixel, angular, or physical unit to convert to.
        '''

        from astropy.table import Table, Column

        tab = Table()

        for name, val, err in zip(self.radprof_parnames, self.radprof_params,
                                  self.radprof_errors):

            # Try converting to the given unit. Assume failures are not length
            # units.
            try:
                conv_val = self._converter.from_pixel(val, unit)
                conv_err = self._converter.from_pixel(err, unit)
            except u.UnitsError:
                conv_val = val
                conv_err = err

            tab[name] = Column(conv_val.reshape((1,)))
            tab[name + "_err"] = Column(conv_err.reshape((1,)))

        # Add on the FWHM
        tab['fwhm'] = Column(self.radprof_fwhm(unit)[0].reshape((1,)))
        tab['fwhm_err'] = Column(self.radprof_fwhm(unit)[1].reshape((1,)))

        # Add on whether the fit was "successful"
        tab['fail_flag'] = Column([self.radprof_fit_fail_flag])

        # Add the type of fit based on the model type
        tab['model_type'] = Column([self.radprof_type])

        return tab

    @property
    def radprof_model(self):
        '''
        The fitted radial profile model.
        '''
        return self._radprof_model

    def plot_radial_profile(self, save_name=None, xunit=u.pix,
                            ax=None):
        '''
        Plot the radial profile of the filament and the fitted model.

        Parameters
        ----------
        xunit : `~astropy.units.Unit`, optional
            Pixel, angular, or physical unit to convert to.
        ax : `~matplotlib.axes`, optional
            Use an existing set of axes to plot the profile.
        '''

        dist, radprof = self.radprofile

        model = self.radprof_model

        conv_dist = self._converter.from_pixel(dist, xunit)

        import matplotlib.pyplot as plt

        if ax is None:
            ax = plt.subplot(111)

        ax.plot(conv_dist, radprof, "kD")
        points = np.linspace(np.min(dist),
                             np.max(dist), 5 * len(dist))
        # Check if units should be kept when evaluating the model
        if not model._supports_unit_fitting:
            points = points.value

        conv_points = np.linspace(np.min(conv_dist),
                                  np.max(conv_dist), 5 * len(conv_dist))
        ax.plot(conv_points, model(points), "r")
        ax.set_xlabel(r'Radial Distance ({})'.format(xunit))
        ax.set_ylabel(r'Intensity ({})'.format(self._yunit))
        ax.grid(True)

        plt.tight_layout()

        if save_name is not None:
            plt.savefig(save_name)

        plt.show()
        if in_ipynb():
            plt.clf()

    def total_intensity(self, bkg_subtract=False, bkg_mod_index=2):
        '''
        Return the sum of all pixels within the FWHM of the filament.

        .. warning::
            `fil_finder_2D` multiplied the total intensity by the angular size
            of a pixel. This function is just the sum of pixel values. Unit
            conversions can be applied on the output if needed.

        Parameters
        ----------
        bkg_subtract : bool, optional
            Subtract off the fitted background level.
        bkg_mod_index : int, optional
            Indicate which element in `Filament2D.radprof_params` is the
            background level. Defaults to 2 for the Gaussian with background
            model.

        Returns
        -------
        total_intensity : `~astropy.units.Quantity`
            The total intensity for the filament.
        '''

        within_fwhm = self._unbin_radprofile[0] <= \
            0.5 * self.radprof_fwhm()[0]
        total_intensity = np.sum(self._unbin_radprofile[1][within_fwhm])

        if bkg_subtract:
            bkg = self.radprof_params[bkg_mod_index]
            if not self.radprof_model._supports_unit_fitting:
                bkg = bkg.value * total_intensity.unit

            total_intensity -= bkg * within_fwhm.sum()

        return total_intensity

    def model_image(self, max_radius=20 * u.pix, bkg_subtract=True,
                    bkg_mod_index=2):
        '''
        Return a model image from the radial profile fit.

        Parameters
        ----------
        max_radius : `~astropy.units.Quantity`, optional
            Set the radius to compute the model to. The outputted array
            will be padded by the number of pixels the max_radius corresponds
            to.
        bkg_subtract : bool, optional
            Subtract off the fitted background level.
        bkg_mod_index : int, optional
            Indicate which element in `Filament2D.radprof_params` is the
            background level. Defaults to 2 for the Gaussian with background
            model.

        Returns
        -------
        model_array : `~astropy.units.Quantity`
            A 2D array computed using the radial profile model.
        '''

        max_radius = self._converter.to_pixel(max_radius).value

        pad_size = int(max_radius)
        skel_arr = self.skeleton(pad_size)

        dists = nd.distance_transform_edt(~skel_arr)
        if self.radprof_model._supports_unit_fitting:
            dists = dists * u.pix

        if not bkg_subtract:
            return self.radprof_model(dists)
        else:
            bkg = self.radprof_params[bkg_mod_index]
            if not self.radprof_model._supports_unit_fitting:
                bkg = bkg.value
            return self.radprof_model(dists) - bkg

    def median_brightness(self, image):
        '''
        Return the median brightness along the skeleton of the filament.

        Parameters
        ----------
        image : `~numpy.ndarray` or `~astropy.units.Quantity`
            The image from which the filament was extracted.

        Returns
        -------
        median_brightness : float or `~astropy.units.Quantity`
            Median brightness along the skeleton.
        '''
        pad_size = 1

        # Do we need to pad the image before slicing?
        input_image = pad_image(image, self.pixel_extents, pad_size)

        skels = self.skeleton(pad_size=pad_size)

        # If the padded image matches the mask size, don't need additional
        # slicing
        if input_image.shape != skels.shape:
            input_image = self.image_slicer(input_image, skels.shape,
                                            pad_size=pad_size)

        assert input_image.shape == skels.shape

        return np.nanmedian(input_image[skels])

    def ridge_profile(self, image):
        '''
        Return the image values along the longest path extent of a filament, or
        from radial slices along the longest path.

        Parameters
        ----------
        image : `~numpy.ndarray` or `~astropy.units.Quantity`
            The image from which the filament was extracted.
        '''

        pad_size = 1

        # Do we need to pad the image before slicing?
        input_image = pad_image(image, self.pixel_extents, pad_size) * \
            u.dimensionless_unscaled

        skels = self.skeleton(pad_size=pad_size, out_type='longpath')

        # If the padded image matches the mask size, don't need additional
        # slicing
        if input_image.shape != skels.shape:
            input_image = self.image_slicer(input_image, skels.shape,
                                            pad_size=pad_size)

        # These should have the same shape now.
        assert input_image.shape == skels.shape

        from .width_profiles.profile_line_width import walk_through_skeleton

        order_pts = walk_through_skeleton(skels)

        if hasattr(image, 'unit'):
            unit = image.unit
        else:
            unit = u.dimensionless_unscaled
            input_image = input_image * unit

        values = []
        for pt in order_pts:
            values.append(input_image[pt[0], pt[1]].value)

        return values * unit

    def profile_analysis(self, image, max_dist=20 * u.pix,
                         num_avg=3, xunit=u.pix):
        '''
        Create profiles of radial slices along the longest path skeleton.
        Profiles created from `~fil_finder.width_profiles.filament_profile`.

        .. note::
            Does not include fitting to the radial profiles. Limited fitting
            of Gaussian profiles is provided in
            `~fil_finder.width_profiles.filament_profile`. See a dedicated
            package like `radfil <https://github.com/catherinezucker/radfil>`_
            for modeling profiles.

        Parameters
        ----------
        image : `~numpy.ndarray` or `~astropy.units.Quantity`
            The image from which the filament was extracted.
        max_dist : astropy Quantity, optional
            The angular or physical (when distance is given) extent to create
            the profile away from the centre skeleton pixel. The entire
            profile will be twice this value (for each side of the profile).
        num_avg : int, optional
            Number of points before and after a pixel that is used when
            computing the normal vector. Using at least three points is
            recommended due to small pixel instabilities in the skeletons.

        Returns
        -------
        dists : `~astropy.units.Quantity`
            Distances in the radial profiles from the skeleton. Units set by
            `xunit`.
        profiles : `~astropy.units.Quantity`
            Radial image profiles.
        '''

        from .width_profiles import filament_profile

        max_dist = self._converter.to_pixel(max_dist)
        pad_size = int(max_dist.value)

        # Do we need to pad the image before slicing?
        input_image = pad_image(image, self.pixel_extents, pad_size)

        if hasattr(image, 'unit'):
            input_image = input_image * image.unit
        else:
            input_image = input_image * u.dimensionless_unscaled

        skels = self.skeleton(pad_size=pad_size, out_type='longpath')

        # If the padded image matches the mask size, don't need additional
        # slicing
        if input_image.shape != skels.shape:
            input_image = self.image_slicer(input_image, skels.shape,
                                            pad_size=pad_size)

        # Check if angular conversions are defined. If not, stay in pixel units
        if hasattr(self._converter, '_ang_size'):
            pixscale = self._converter.to_angular(1 * u.pix)
            ang_conv = True
        else:
            pixscale = 1.0 * u.deg
            ang_conv = False

        dists, profiles = filament_profile(skels, input_image, pixscale,
                                           max_dist=max_dist,
                                           distance=None,
                                           fit_profiles=False,
                                           bright_unit=input_image.unit)

        # First put the distances into pixel units
        if ang_conv:
            dists = [self._converter.to_pixel(dist) for dist in dists]
        else:
            # Already in pixel units.
            dists = [dist.value * u.pix for dist in dists]

        # Convert the distance units
        dists = [self._converter.from_pixel(dist, xunit) for dist in dists]

        return dists, profiles

    def radprof_table(self, xunit=u.pix):
        '''
        Return the radial profile as a table.

        Parameters
        ----------
        xunit : `~astropy.units.Unit`, optional
            Spatial unit to convert radial profile distances.

        Returns
        -------
        tab : `~astropy.table.Table`
            Table with the radial profile distance and values.
        '''
        from astropy.table import Column, Table

        dists = Column(self._converter.from_pixel(self._radprofile[0], xunit))
        vals = Column(self._radprofile[1])

        tab = Table()
        tab['distance'] = dists
        tab['values'] = vals

        return tab

    def branch_table(self, include_rht=False):
        '''
        Save the branch properties of the filament.

        Parameters
        ----------
        include_rht : bool, optional
            If `branches=True` is used in `Filament2D.exec_rht`, the branch
            orientation and curvature will be added to the table.

        Returns
        -------
        tab : `~astropy.table.Table`
            Table with the branch properties.
        '''

        from astropy.table import Table, Column

        branch_data = self.branch_properties.copy()
        del branch_data['pixels']
        del branch_data['number']

        if include_rht:
            branch_data['orientation'] = self.orientation_branches
            branch_data['curvature'] = self.curvature_branches

        tab = Table([Column(branch_data[key]) for key in branch_data],
                    names=branch_data.keys())
        return tab

    def save_fits(self, savename, image, pad_size=20 * u.pix, header=None,
                  **model_kwargs):
        '''
        Save a stamp of the image centered on the filament, the skeleton,
        the longest path skeleton, and the model.

        Parameters
        ----------
        image : `~numpy.ndarray` or `~astropy.units.Quantity`
            The image from which the filament was extracted.
        pad_size : `~astropy.units.Quantity`, optional
            Size to pad the saved arrays by.
        header : `~astropy.io.fits.Header`, optional
            Provide a FITS header to save to. If `~Filament2D` was
            given WCS information, this will be used if no header is given.
        model_kwargs : Passed to `~Filament2D.model_image`.

        '''

        pad_size = int(self._converter.to_pixel(pad_size).value)

        # Do we need to pad the image before slicing?
        input_image = pad_image(image, self.pixel_extents, pad_size)

        skels = self.skeleton(pad_size=pad_size, out_type='all')
        skels_lp = self.skeleton(pad_size=pad_size, out_type='longpath')

        # If the padded image matches the mask size, don't need additional
        # slicing
        if input_image.shape != skels.shape:
            input_image = self.image_slicer(input_image, skels.shape,
                                            pad_size=pad_size)

        model = self.model_image(max_radius=pad_size * u.pix,
                                 **model_kwargs)
        if hasattr(model, 'unit'):
            model = model.value

        from astropy.io import fits
        import time

        if header is None:
            if hasattr(self._converter, "_wcs"):
                header = self._converter._wcs.to_header()
            else:
                header = fits.Header()

        # Strip off units if the image is a Quantity
        if hasattr(input_image, 'unit'):
            input_image = input_image.value.copy()

        hdu = fits.PrimaryHDU(input_image, header)

        skel_hdr = header.copy()
        skel_hdr['BUNIT'] = ("", "bool")
        skel_hdr['COMMENT'] = "Skeleton created by fil_finder on " + \
            time.strftime("%c")

        skel_hdu = fits.ImageHDU(skels.astype(int), skel_hdr)

        skel_lp_hdu = fits.ImageHDU(skels_lp.astype(int), skel_hdr)

        model_hdu = fits.ImageHDU(model, header)

        hdulist = fits.HDUList([hdu, skel_hdu, skel_lp_hdu, model_hdu])

        hdulist.writeto(savename)

    def to_pickle(self, savename):
        '''
        Save a Filament2D class as a pickle file.

        Parameters
        ----------
        savename : str
            Name of the pickle file.
        '''

        with open(savename, 'wb') as output:
                pickle.dump(self, output, -1)

    @staticmethod
    def from_pickle(filename):
        '''
        Load a Filament2D from a pickle file.

        Parameters
        ----------
        filename : str
            Name of the pickle file.
        '''
        with open(filename, 'rb') as input:
                self = pickle.load(input)

        return self


class Filament3D(FilamentNDBase):
    """docstring for Filament3D"""
    def __init__(self, arg):
        super(Filament3D, self).__init__()
        self.arg = arg




# Licensed under an MIT open source license - see LICENSE

"""
**THESE ARE NOT CURRENTLY USED IN THE ALGORITHM**

Finding/Analyzing Cores Routines for fil-finder

Contains: abs_thresh
subtract_cores

"""

# from .utilities import *

# import numpy as np
# from scipy.ndimage import label


# def abs_thresh(arr,thresh_value,img_scale,img_freq):
#     arr = np.array(arr)

#     conversion = planck(20.,img_freq)*1e20*20. * (2.32 * 1.67e-30)*img_scale**-1 #Converted to MJy/sr
#     core_thresh = conversion * thresh_value

#     core_array = arr > core_thresh

#     return core_array

# def subtract_cores(core_array,full_output=False):

#     lab_core,num_core = label(core_array,eight_con())

#     for n in range(num_core):
#         x,y = np.where(lab_core==n+1)

#     if len(x)>10:
#         filarr = core_array[min(x):max(x),min(y):max(y)]

#     try:
#         fit,err = fit2dgaussian(filarr)
#         for i in range(len(x)):
#             core_array[x[i],y[i]] = fit[-1]
#     ## Set Core region to fit background, should be approx of rest of filament

#     except:
#         print("Fit Failed on Core %s" % (n))


#     if full_output:
#         return core_array, fit, err
#     else:
#         return core_array


# if __name__ == "__main__":
#     import sys
#     fib(int(sys.argv[1]))


# Licensed under an MIT open source license - see LICENSE

from .utilities import find_nearest

import numpy as np
import scipy.ndimage as nd
import scipy.optimize as op
from scipy.integrate import quad
from scipy.stats import scoreatpercentile, percentileofscore
from scipy.interpolate import interp1d
from scipy.signal import argrelmax, argrelmin
from warnings import warn
from astropy.modeling import fitting, models
import astropy.modeling as mod


def dist_transform(labelisofil, filclean_all):
    '''

    Recombines the cleaned skeletons from final analysis and takes the
    Euclidean Distance Transform of each. Since each filament is in an
    array defined by its own size, the offsets need to be taken into account
    when adding back into the master array.

    Parameters
    ----------
    labelisofil : list
        Contains arrays of the cleaned individual skeletons
    filclean_all : numpy.ndarray
        Master array with all final skeletons.

    Returns
    -------
    dist_transform_all : numpy.ndarray
        A Euclidean Distance Transform of all of the skeletons combined.
    dist_transform_sep : list
        Contains the Euclidean Distance Transform of each
        individual skeleton.
    '''

    dist_transform_sep = []

    for skel_arr in labelisofil:
        if np.max(skel_arr) > 1:
            skel_arr[np.where(skel_arr > 1)] = 1
        dist_transform_sep.append(
            nd.distance_transform_edt(np.logical_not(skel_arr)))

    # Distance Transform of all cleaned filaments
    dist_transform_all = nd.distance_transform_edt(
        np.logical_not(filclean_all))

    return dist_transform_all, dist_transform_sep


def gaussian_model(dist, radprof, with_bkg=True):
    '''
    Return a Gaussian model with initial parameter guesses. The model is for
    radial profiles and the peak is assumed to be fixed at the centre.

    Parameters
    ----------
    dist : `~numpy.ndarray`
        Distance bins of the radial profile.
    radprof : `~numpy.ndarray`
        The binned radial profile.
    with_bkg : bool, optional
        Add constant background to the fit when enabled.
    '''

    amp = radprof.max()
    wid_amp = amp / np.exp(0.5)
    idx = np.abs(radprof - wid_amp).argmin()
    inner_width = dist[idx]

    mod = models.Gaussian1D()
    # Check if the version of astropy supports units in the model.
    if mod._supports_unit_fitting:
        mod.amplitude = amp
        mod.mean = 0.0 * inner_width.unit
        mod.stddev = inner_width
    else:
        mod.amplitude = amp.value
        mod.mean = 0.0
        mod.stddev = inner_width.value

    # Fix the mean to 0, since this is for radial profiles.
    mod.mean.fixed = True

    if with_bkg:
        bkg_mod = models.Const1D()
        if bkg_mod._supports_unit_fitting:
            bkg_mod.amplitude = np.min(radprof)
        else:
            bkg_mod.amplitude = np.min(radprof).value
        bkg_mod = bkg_mod.rename("Bkg")
        mod = mod + bkg_mod

        # Check if the compound model can handle units?
        if not mod._supports_unit_fitting:
            # Strip the units out. Hopefully I can get rid of this in a
            # future release...
            mod_new = models.Gaussian1D() + models.Const1D()

            mod_new.amplitude_0 = mod.amplitude_0.value
            mod_new.mean_0 = mod.mean_0.value
            mod_new.mean_0.fixed = True
            mod_new.stddev_0 = mod.stddev_0.value
            mod_new.amplitude_1 = mod.amplitude_1.value

            mod = mod_new

    return mod


def fit_radial_model(dist, radprof, model, fitter=None, weights=None,
                     **fitter_kwargs):
    '''
    Fit a model to the radial profile.

    Parameters
    ----------
    dist :

    radprof :

    model : `~astropy.modeling.models.Fittable1DModel`
        An astropy model object to fit to.
    fitter : `~astropy.modeling.fitting.Fitter`, optional
        One of the fitting classes from astropy. This should probably be a
        non-linear fitting algorithm, but it depends on the chosen model.
        Defaults to a Levenberg-Marquardt fitter.
    weights : `~numpy.ndarray`, optional
        Weights to apply in the fitting.
    '''

    if not isinstance(model, mod.Model):
        raise TypeError("The model must be a 1D astropy model.")

    if fitter is None:
        fitter = fitting.LevMarLSQFitter()
    else:
        if not isinstance(fitter, fitting.Fitter):
            raise TypeError("The fitter must be one of the "
                            "astropy.modeling.fitting classes.")

    if model._supports_unit_fitting:
        xdata = dist
        ydata = radprof
    else:
        xdata = dist.value
        ydata = radprof.value

    fitted_mod = fitter(model, xdata, ydata, weights=weights, **fitter_kwargs)

    return fitted_mod, fitter


def gauss_model(distance, rad_profile, weights, img_beam):
    '''
    Fits a Gaussian to the radial profile of each filament.

    Parameters
    ----------
    distance : list
        Distances from the skeleton.
    rad_profile : list
        Intensity values from the image.
    weights : list
        Weights to be used for the fit.
    img_beam : float
        FWHM of the beam size.

    Returns
    -------
    fit : numpy.ndarray
        Fit values.
    fit_errors : numpy.ndarray
        Fit errors.
    gaussian : function
        Function used to fit the profile.
    parameters : list
        Names of the parameters.
    fail_flag : bool
        Identifies a failed fit.
    '''

    p0 = (np.max(rad_profile), np.std(distance), np.min(rad_profile))
    parameters = ["Amplitude", "Width", "Background", "FWHM"]

    def gaussian(x, *p):
        '''
        Parameters
        ----------
        x : list or numpy.ndarray
            1D array of values where the model is evaluated
        p : tuple
            Components are:
            * p[0] Amplitude
            * p[1] Width
            * p[2] Background
        '''
        return (p[0] - p[2]) * np.exp(-1 * np.power(x, 2) /
                                      (2 * np.power(p[1], 2))) + p[2]

    try:
        fit, cov = op.curve_fit(gaussian, distance, rad_profile, p0=p0,
                                maxfev=100 * (len(distance) + 1), sigma=weights)
        fit_errors = np.sqrt(np.diag(cov))
    except:
        print("curve_fit failed.")
        fit, fit_errors = p0, None
        return fit, fit_errors, gaussian, parameters, True

    # Because of how function is defined,
    # fit function can get stuck at negative widths
    # This doesn't change the output though.
    fit[1] = np.abs(fit[1])

    # Deconvolve the width with the beam size.
    factor = 2 * np.sqrt(2 * np.log(2))  # FWHM factor
    deconv = (factor * fit[1]) ** 2. - img_beam ** 2.
    if deconv > 0:
        fit_errors = np.append(
            fit_errors, (factor**2 * fit[1] * fit_errors[1]) / np.sqrt(deconv))
        fit = np.append(fit, np.sqrt(deconv))
    else:  # Set to zero, can't be deconvolved
        fit = np.append(fit, 0.0)
        fit_errors = np.append(fit_errors, 0.0)

    fail_flag = False
    # The bkg level may be quite uncertain. Only look for poor amplitude or
    # width constraints.
    fail_conditions = fit_errors is None or \
        fit[0] < fit[2] or (fit_errors[:2] > np.abs(fit[:2])).any()
    if fail_conditions:
        fail_flag = True

    return fit, fit_errors, gaussian, parameters, fail_flag


def nonparam_width(distance, rad_profile, unbin_dist, unbin_prof,
                   img_beam=None, bkg_percent=5, peak_percent=99):
    '''
    Estimate the width and peak brightness of a filament non-parametrically.
    The intensity at the peak and background is estimated. The profile is then
    interpolated over in order to find the distance corresponding to these
    intensities. The width is then estimated by finding the distance where
    the intensity drops to 1/e.

    Parameters
    ----------
    distance : list
        Binned distances from the skeleton.
    rad_profile : list
        Binned intensity values from the image.
    unbin_dist : list
        Unbinned distances.
    unbin_prof : list
        Unbinned intensity values.
    img_beam : float, optional
        FWHM of the beam size.
    bkg_percent : float, optional
        Percentile of the data to estimate the background.
    peak_percent : float, optional
        Percentile of the data to estimate the peak of the profile.

    Returns
    -------
    params : numpy.ndarray
        Estimated parameter values.
    param_errors : numpy.ndarray
        Estimated errors.
    fail_flag : bool
        Indicates whether the fit failed.
    '''

    fail_flag = False

    # Find the intensities at the given percentiles
    bkg_intens = scoreatpercentile(rad_profile, bkg_percent)
    peak_intens = scoreatpercentile(rad_profile, peak_percent)

    # Interpolate over the bins in distance
    interp_bins = np.linspace(0.0, np.max(distance), 10 * len(distance))
    interp_profile = np.interp(interp_bins, distance, rad_profile)

    # Find the width by looking for where the intensity drops to 1/e from the
    # peak
    target_intensity = ((peak_intens - bkg_intens) / np.exp(0.5)) + bkg_intens
    width = interp_bins[np.where(interp_profile ==
                                 find_nearest(interp_profile, target_intensity))][0]

    # Estimate the width error by looking +/-5 percentile around the target
    # intensity
    target_percentile = percentileofscore(rad_profile, target_intensity)
    upper = scoreatpercentile(
        rad_profile, np.min((100, target_percentile + 5)))
    lower = scoreatpercentile(rad_profile, np.max((0, target_percentile - 5)))

    error_range_pts = np.logical_and(unbin_prof > lower, unbin_prof < upper)
    width_error = np.max(unbin_dist[error_range_pts]) -\
        np.min(unbin_dist[error_range_pts])

    # Deconvolve the width with the beam size.
    factor = 2 * np.sqrt(2 * np.log(2))  # FWHM factor
    if img_beam is not None:

        deconv = (width * factor) ** 2. - img_beam ** 2.
        if deconv > 0:
            fwhm_width = np.sqrt(deconv)
            fwhm_error = (factor**2 * width * width_error) / fwhm_width
        else:  # Set to zero, can't be deconvolved
            # If you can't devolve it, set it to minimum, which is the
            # beam-size.
            fwhm_width = 0.0
            fwhm_error = 0.0
    else:
        fwhm_width = width * factor
        fwhm_error = width_error * factor

    # Check where the "background" and "peak" are. If the peak distance is
    # greater, we are simply looking at a bad radial profile.
    bkg_dist = np.median(
        interp_bins[np.where(interp_profile ==
                             find_nearest(interp_profile, bkg_intens))])
    peak_dist = np.median(
        interp_bins[np.where(interp_profile ==
                             find_nearest(interp_profile, peak_intens))])
    bkg_error = np.std(
        unbin_prof[unbin_dist >= find_nearest(unbin_dist, bkg_dist)])
    peak_error = np.std(
        unbin_prof[unbin_dist <= find_nearest(unbin_dist, peak_dist)])

    # Check if there are unrealistic estimates
    if peak_dist > bkg_dist or fwhm_error > fwhm_width:
        fail_flag = True

    params = np.array([peak_intens, width, bkg_intens, fwhm_width])
    param_errors = np.array([peak_error, width_error, bkg_error, fwhm_error])

    return params, param_errors, fail_flag


def radial_profile(img, dist_transform_all, dist_transform_sep, offsets,
                   img_scale=1.0, bins=None, bintype="linear",
                   weighting="number", return_unbinned=True, auto_cut=False,
                   pad_to_distance=0.15, max_distance=0.3, auto_cut_kwargs={},
                   debug_mode=False):
    '''
    Fits the radial profiles to all filaments in the image.

    Parameters
    ----------
    img : numpy.ndarray
              The original image.
    dist_transform_all : numpy.ndarray
        The distance transform of all the skeletons.
    dist_transform_sep : list
        The distance transforms of each individual skeleton.
    offsets : list
        Contains the indices where each skeleton was cut out of the original
        array.
    img_scale : float
        Pixel to physical scale conversion.
    bins : numpy.ndarray, optional
        Bins to use for the profile fitting.
    bintype : str, optional
        "linear" for linearly spaced bins; "log" for log-spaced bins.
        Default is "linear".
    weighting : str, optional
        "number" is by the number of points in each bin; "var" is the
        variance of the values in each bin. Default is "number".
    return_unbinned : bool
        If True, returns the unbinned data as well as the binned.
    auto_cut : bool, optional
        Enables the auto cutting routines.
    pad_to_distance : float, optional
        Include pixels in the profile whose distance in dist_transform_sep
        is less than dist_transform_all + pad_to_distance. This is useful
        for creating profiles in crowded regions. If set to 0.0, no padding
        is done. Must be less than max_distance.
    max_distance : float, optional
        Cuts the profile at the specified physical distance (in pc).
    debug_mode : bool, optional
        Enables plotting of which pixels are being used in the radial profile.

    Returns
    -------
    bin_centers : numpy.ndarray
        Center of the bins used in physical units.
    radial_prof : numpy.ndarray
        Binned intensity profile.
    weights : numpy.ndarray
        Weights evaluated for each bin.
    '''

    if max_distance <= 0.0:
        raise ValueError("max_distance must be positive.")

    if pad_to_distance < 0.0 or pad_to_distance > max_distance:
        raise ValueError("pad_to_distance must be positive and less than "
                         "max_distance.")

    width_value = []
    width_distance = []
    x, y = np.where(np.isfinite(dist_transform_sep) *
                    (dist_transform_sep <= max_distance / img_scale))
    # Transform into coordinates of master image
    # Originally had a +1 offset, but new code doesn't need this
    # Correction added directly into fil_finder_2D
    x_full = x + offsets[0][0]  # - 1
    y_full = y + offsets[0][1]  # - 1

    pad_pixel_distance = int(pad_to_distance * img_scale ** -1)

    # Don't necessarily need dist_transform_all. If None, skip some parts
    if dist_transform_all is None:
        check_global = False
    else:
        check_global = True

    # Check if the image has a unit
    if hasattr(img, 'unit'):
        img_vals = img.value
    else:
        img_vals = img

    valids = np.zeros_like(dist_transform_sep, dtype=bool)

    for i in range(len(x)):
        # Check overall distance transform to make sure pixel belongs to proper
        # filament
        img_val = img_vals[x_full[i], y_full[i]]
        sep_dist = dist_transform_sep[x[i], y[i]]
        if check_global:
            glob_dist = dist_transform_all[x_full[i], y_full[i]]
        if np.isfinite(img_val):
            if check_global:
                # Include the point if it falls within the pad distance.
                if sep_dist <= glob_dist + pad_pixel_distance:
                    width_value.append(img_val)
                    width_distance.append(sep_dist)
                    # valids[x_full[i], y_full[i]] = True
                    valids[x[i], y[i]] = True
            else:
                width_value.append(img_val)
                width_distance.append(sep_dist)
                # valids[x_full[i], y_full[i]] = True
                valids[x[i], y[i]] = True

    if debug_mode:
        import matplotlib.pyplot as plt
        plt.subplot(121)
        plt.imshow(dist_transform_sep)
        plt.contour(valids, colors='g')
        plt.contour(dist_transform_all == 0., colors='m')
        plt.contour(dist_transform_sep == 0., colors='c')
        plt.subplot(122)
        plt.imshow(dist_transform_all)
        try:
            plt.contour(dist_transform_sep <= max_distance / img_scale,
                        colors='b')
        except ValueError:
            print("No contour")
        plt.draw()
        raw_input("?")
        plt.clf()

    if len(width_distance) == 0:
        warn("No valid pixels for radial profile found.")
        return None

    width_value = np.asarray(width_value)
    width_distance = np.asarray(width_distance)

    # Binning
    if bins is None:
        nbins = int(np.sqrt(len(width_value)))
        maxbin = np.max(width_distance)
        if bintype is "log":
            # bins must start at 1 if logspaced
            bins = np.logspace(0, np.log10(maxbin), nbins + 1)
        elif bintype is "linear":
            bins = np.linspace(0, maxbin, nbins + 1)

    whichbins = np.digitize(width_distance, bins)
    bin_centers = (bins[1:] + bins[:-1]) / 2.0

    radial_prof = np.zeros_like(bin_centers)
    weights = np.zeros_like(bin_centers)

    for nbin in range(1, int(nbins) + 1):

        bin_posns = whichbins == nbin

        # Skip any empty bins
        if bin_posns.sum() == 0:
            continue

        radial_prof[nbin - 1] = np.median(width_value[bin_posns])

        if weighting == "number":
            weights[nbin - 1] = whichbins[bin_posns].sum()
        elif weighting == "var":
            weights[nbin - 1] = np.nanvar(width_value[bin_posns])

    # Remove all empty bins
    radial_prof = radial_prof[weights > 0]
    bin_centers = bin_centers[weights > 0]
    weights = weights[weights > 0]

    # Put bins in the physical scale.
    bin_centers *= img_scale
    width_distance *= img_scale

    if auto_cut:
        bin_centers, radial_prof, weights = \
            _smooth_and_cut(bin_centers, radial_prof, weights,
                            **auto_cut_kwargs)

    if return_unbinned:
        width_distance = width_distance[np.isfinite(width_value)]
        width_value = width_value[np.isfinite(width_value)]
        return bin_centers, radial_prof, weights, width_distance, width_value
    else:
        return bin_centers, radial_prof, weights


def _smooth_and_cut(bins, values, weights, interp_factor=10,
                    pad_cut=5, smooth_size=0.03, min_width=0.1):
    '''
    Smooth the radial profile and cut if it increases at increasing
    distance. Also checks for profiles with a plateau between two decreasing
    profiles and cut out the last one (as it is not the local profile).

    Parameters
    ----------
    bins : numpy.ndarray
        Bins for the profile.
    values : numpy.ndarray
        Values in each bin.
    weights : numpy.ndarray
        Weights for each bin. These are only clipped to the same position as
        the rest of the profile. Otherwise, no alteration is made.
    interp_factor : int, optional
        The factor to increase the number of bins by for interpolation.
    pad_cut : int, optional
        Add additional bins after the cut is found. The smoothing often cuts
        out some bins which follow the desired profile.
    smooth_size : float, optional
        Set the smoothing size when finding local extrema. The value should
        have the same units as given in `bins`.
    min_width : float, optional
        Ignore local minima below this minimum width.

    Returns
    -------
    cut_bins : numpy.ndarray
        Bins for the profile with a possible cutoff.
    cut_values : numpy.ndarray
        Values in each bin with a possible cutoff.
    cut_weights : numpy.ndarray
        Weights for each bin with a possible cutoff.

    '''

    # Interpolate the points onto a finer spacing
    smooth_bins = np.linspace(bins.min(), bins.max(),
                              interp_factor * bins.size)
    smooth_bin_width = smooth_bins[1] - smooth_bins[0]

    smooth_val = interp1d(bins, values, kind='cubic')(smooth_bins)

    # Adjust size based on interpolation upsample factor
    window_size = int(np.floor(smooth_size / smooth_bin_width))
    # Must be odd!
    if window_size % 2 == 0:
        window_size -= 1

    # Perform a moving average on the interpolated points.
    pad_add = int((window_size - 1) // 2)
    smooth_val = \
        np.convolve(smooth_val, np.ones((window_size,)) / window_size,
                    mode='valid')

    grad = np.gradient(smooth_val, smooth_bins[1] - smooth_bins[0])

    # The shapes need equal so the index to cut at is correct.
    assert smooth_bins.size == smooth_val.size + 2 * pad_add

    cut = crossings_nonzero_all(grad) + pad_add

    # Check for evidence of second drop-off
    new_cut = None

    # Look for local max and mins (must hold True for range of ~0.05 pc)
    bin_diff = smooth_bins[1] - smooth_bins[0]
    loc_mins = argrelmin(grad, order=int(smooth_size / bin_diff))[0] + pad_add
    loc_maxs = argrelmax(grad, order=int(smooth_size / bin_diff))[0] + pad_add

    # Discard below some minimum width (defaults to 0.1 pc).
    loc_mins = loc_mins[smooth_bins[loc_mins] > min_width]
    loc_maxs = loc_maxs

    if loc_mins.size > 0 and loc_maxs.size > 0:
        i = 0
        while True:
            loc_min = loc_mins[i]

            difference = loc_min - loc_maxs
            if (difference > 0).any():
                new_cut = loc_maxs[np.argmin(difference[difference > 0])]
                if smooth_bins[new_cut] > min_width:
                    break

            i += 1

            if i == loc_mins.size:
                break

    if new_cut == 0:
        new_cut = None

    # No cut is found, so no values are sliced off.
    if cut.size == 0 and new_cut is None:
        cut_posn = bins.size
    # Only a plateau cut was found.
    elif cut.size == 0 and new_cut is not None:
        cut_posn = _nearest_idx(bins, smooth_bins[new_cut])
    else:
        # Check for a cut given by the plateau check.
        # If there is isn't one or it is at a greater radius, use the first
        # zero gradient crossing.
        if new_cut is None or new_cut >= cut[0]:
            cut_used = cut[0]
        else:
            cut_used = new_cut

        cut_posn = _nearest_idx(bins, smooth_bins[cut_used])

    # Now adjust by the given pad_cut size
    # Always use +1 to cut to the extrema point.
    cut_posn += 1 + pad_cut

    cut_bins = bins[:cut_posn]
    cut_vals = values[:cut_posn]
    cut_weights = weights[:cut_posn]

    return cut_bins, cut_vals, cut_weights


def _nearest_idx(array, value):
    return (np.abs(array - value)).argmin()


def crossings_nonzero_all(data):
    assert isinstance(data, np.ndarray)
    pos = data > 0
    npos = ~pos
    return ((pos[:-1] & npos[1:]) | (npos[:-1] & pos[1:])).nonzero()[0]



# Licensed under an MIT open source license - see LICENSE

"""
Utility functions for fil-finder package


"""

import itertools
import numpy as np


def find_nearest(array, value):
    idx = (np.abs(array - value)).argmin()
    return array[idx]

##########################################################################
# Simple fcns used throughout module
##########################################################################


def eight_con():
    return np.ones((3, 3))


def distance(x, x1, y, y1):
    return np.sqrt((x - x1) ** 2.0 + (y - y1) ** 2.0)


def pad_image(image, extents, pad_size, constant=0):
    '''
    Figure out where an image needs to be padded based on pixel extents.
    '''

    # Now we look for where the extents with padding will run into a boundary
    pad_vector = []
    for i, extent in enumerate(zip(*extents)):

        lower = extent[0] - pad_size
        if lower > 0:
            lower = 0

        # First axis upper
        upper = extent[1] + pad_size - image.shape[i] + 1
        if upper < 0:
            upper = 0

        pad_vector.append((-lower, upper))

    return np.pad(image, pad_vector, mode='constant', constant_values=constant)


def shifter(l, n):
    return l[n:] + l[:n]


def product_gen(n):
    for r in itertools.count(1):
        for i in itertools.product(n, repeat=r):
            yield "".join(i)


def red_chisq(data, fit, nparam, sd):
    N = data.shape[0]
    stat = np.sum(((fit - data) / sd) ** 2.) / float(N - nparam - 1)
    if hasattr(stat, 'unit'):
        stat = stat.value
    return stat


def in_ipynb():
    try:
        cfg = get_ipython().config
        if cfg['IPKernelApp']['parent_appname'] == 'ipython-notebook':
            return True
        else:
            return False
    except NameError:
        return False


def round_to_odd(x):
    return int((np.ceil((np.ceil(x) / 2) + 0.5) * 2) - 1)


def threshold_local(image, *args, **kwargs):
    '''
    skimage changed threshold_adaptive to threshold_local. This wraps both
    to ensure the same behaviour with old and new versions.
    '''
    try:
        from skimage.filters import threshold_local
        mask = image > threshold_local(image, *args, **kwargs)
    except ImportError:
        from skimage.filters import threshold_adaptive
        mask = threshold_adaptive(image, *args, **kwargs)

    return mask



# Licensed under an MIT open source license - see LICENSE

from ._astropy_init import *

if not _ASTROPY_SETUP_:
    from .filfind_class import fil_finder_2D
    from .filfinder2D import FilFinder2D
    from .filament import FilamentNDBase, Filament2D
    from .width_profiles import filament_profile



# Licensed under an MIT open source license - see LICENSE

from astropy.io.fits import PrimaryHDU
from astropy.utils.decorators import format_doc
import numpy as np
import astropy.units as u
import re
import warnings

try:
    from spectral_cube import Projection, Slice
    SPECTRALCUBE_INSTALL = True
except ImportError:
    SPECTRALCUBE_INSTALL = False


allowed_types = ["numpy.ndarray", "astropy.io.fits.PrimaryHDU",
                 "spectral_cube.Projection", 'spectral_cube.Slice',
                 "astropy.units.Quantity"]

input_doc = \
    '''
    Accept a variety of input data forms and return those expected by the
    various statistics.

    Parameters
    ----------
    data : {}
        Data to be used with a given statistic or distance metric. no_header
        must be enabled when passing only an array in.
    header : `astropy.io.fits.Header`, optional
        Pass a header when data is a numpy array or `astropy.units.Quantity`.

    Returns
    -------
    ouput_data : tuple or np.ndarray
        A tuple containing the data and the header. Or an array when no_header
        is enabled.
    '''.format(" or ".join(allowed_types))


@format_doc(input_doc)
def input_data(data, header=None):
    '''
    '''

    output_data = None

    if SPECTRALCUBE_INSTALL:
        if isinstance(data, Projection) or isinstance(data, Slice):
            # spectral-cube has dimensionality checks
            output_data = {'data': data.quantity, 'header': data.header}

    if output_data is None:
        if isinstance(data, PrimaryHDU):
            squeeze_data = data.data.squeeze()
            if 'BUNIT' in data.header:
                unit = convert_bunit(data.header['BUNIT'])
            else:
                unit = u.dimensionless_unscaled
            output_data = {"data": squeeze_data * unit,
                           "header": data.header}

        elif isinstance(data, np.ndarray) or isinstance(data, u.Quantity):
            squeeze_data = data.squeeze()

            output_data = {}

            if header is not None:
                if not hasattr(data, 'unit'):
                    # Attach the BUNIT if defined
                    if 'BUNIT' in header:
                        bunit = convert_bunit(header['BUNIT'])
                        squeeze_data = squeeze_data * bunit

                else:
                    # Make sure the BUNIT in the header is the same as the data
                    header = header.copy()
                    header['BUNIT'] = data.unit.to_string()

                output_data['header'] = header

            if not hasattr(data, 'unit'):
                unit = u.dimensionless_unscaled
                squeeze_data = squeeze_data * unit

            output_data["data"] = squeeze_data
        else:
            raise TypeError("Input data is not of an accepted form:"
                            "{}".format(allowed_types))

    if not dim_check(output_data["data"]):
        raise TypeError("Data must be 2D. Please re-check the inputs.")

    return output_data


def dim_check(arr):
    return True if arr.ndim == 2 else False


def convert_bunit(bunit):
    '''
    Convert a BUNIT string to a quantity

    Stolen from spectral-cube: https://github.com/radio-astro-tools/spectral-cube/blob/c6ef0e6b1f8ba54116c8f558466aff4efb8caf60/spectral_cube/cube_utils.py#L423

    Parameters
    ----------
    bunit : str
        String to convert to an `~astropy.units.Unit`

    Returns
    -------
    unit : `~astropy.unit.Unit`
        Corresponding unit.
    '''

    # special case: CASA (sometimes) makes non-FITS-compliant jy/beam headers
    bunit_lower = re.sub("\s", "", bunit.lower())
    if bunit_lower == 'jy/beam':
        unit = u.Jy / u.beam
    else:
        try:
            unit = u.Unit(bunit)
        except ValueError:
            warnings.warn("Could not parse unit {0}".format(bunit))
            unit = None

    return unit




import pytest

from ..width import nonparam_width, gauss_model, radial_profile
from .testing_utils import generate_filament_model

import numpy as np
import numpy.testing as npt
from scipy import ndimage as nd


def generate_gaussian_profile(pts, width=3.0, amplitude=2.0, background=0.5):
    return amplitude * np.exp(- pts ** 2 / (2 * width ** 2)) + background


def test_nonparam():

    pts = np.linspace(0, 10, 100)

    profile = generate_gaussian_profile(pts)

    params, errors, fail = \
        nonparam_width(pts, profile, pts, profile, 1.0, 5, 99)

    # This shouldn't be failing
    assert fail is False

    # Check the amplitude
    npt.assert_allclose(params[0], 2.5, atol=0.01)
    # Width
    npt.assert_allclose(params[1], 3.0, atol=0.01)
    # Background
    npt.assert_allclose(params[2], 0.5, atol=0.02)


def test_gaussian():

    pts = np.linspace(0, 10, 100)

    profile = generate_gaussian_profile(pts)

    params, errors, _, _, fail = \
        gauss_model(pts, profile, np.ones_like(pts), 1.0)

    # Check the amplitude
    npt.assert_allclose(params[0], 2.5, atol=0.01)
    # Width
    npt.assert_allclose(params[1], 3.0, atol=0.01)
    # Background
    npt.assert_allclose(params[2], 0.5, atol=0.02)


@pytest.mark.parametrize(('theta'), [(0.0)])
def test_radial_profile_output(theta):

    model, skeleton = generate_filament_model(width=10.0,
                                              amplitude=1.0, background=0.0)

    dist_transform = nd.distance_transform_edt((~skeleton).astype(np.int))

    dist, radprof, weights, unbin_dist, unbin_radprof = \
        radial_profile(model, dist_transform, dist_transform,
                       ((0, 0), (model.shape[0] // 2, model.shape[1] // 2)),
                       img_scale=1.0, auto_cut=False, max_distance=20)

    params, errors, _, _, fail = \
        gauss_model(dist, radprof, np.ones_like(dist), 1.0)

    npt.assert_allclose(params[:-1], [1.0, 10.0, 0.0], atol=1e-1)


@pytest.mark.parametrize(('cutoff'), [(10.0), (20.0), (30.0)])
def test_radial_profile_cutoff(cutoff):

    model, skeleton = generate_filament_model(width=10.0,
                                              amplitude=1.0, background=0.0)

    dist_transform = nd.distance_transform_edt((~skeleton).astype(np.int))

    dist, radprof, weights, unbin_dist, unbin_radprof = \
        radial_profile(model, dist_transform, dist_transform,
                       ((0, 0), (model.shape[0] // 2, model.shape[1] // 2)),
                       img_scale=1.0, auto_cut=False, max_distance=cutoff)

    assert unbin_dist.max() == cutoff
    assert dist.max() < cutoff


@pytest.mark.parametrize(('padding'), [(5.0), (10.0), (20.0)])
def test_radial_profile_padding(padding, max_distance=20.0):

    model, skeleton = generate_filament_model(width=10.0,
                                              amplitude=1.0, background=0.0)

    dist_transform = nd.distance_transform_edt((~skeleton).astype(np.int))

    dist, radprof, weights, unbin_dist, unbin_radprof = \
        radial_profile(model, dist_transform, dist_transform,
                       ((0, 0), (model.shape[0] // 2, model.shape[1] // 2)),
                       img_scale=1.0, auto_cut=False,
                       max_distance=max_distance, pad_to_distance=padding)

    if padding <= max_distance:
        assert unbin_dist.max() == max_distance
        assert dist.max() < max_distance
    else:
        assert unbin_dist.max() == padding
        assert dist.max() < padding


@pytest.mark.xfail(raises=ValueError)
def test_radial_profile_fail_pad(padding=30.0, max_distance=20.0):
    '''
    Cannot pad greater than max_distance
    '''
    model, skeleton = generate_filament_model(width=10.0,
                                              amplitude=1.0, background=0.0)

    dist_transform = nd.distance_transform_edt((~skeleton).astype(np.int))

    dist, radprof, weights, unbin_dist, unbin_radprof = \
        radial_profile(model, dist_transform, dist_transform,
                       ((0, 0), (model.shape[0] // 2, model.shape[1] // 2)),
                       img_scale=1.0, auto_cut=False,
                       max_distance=max_distance, pad_to_distance=padding)


def test_radial_profile_autocut():
    '''
    Test auto-cutting with a secondary offset peak.
    '''

    model, skeleton = generate_filament_model(width=10.0,
                                              amplitude=1.0, background=0.0)

    model += np.roll(model, -30, axis=0).copy()
    model += np.roll(model, +30, axis=0).copy()

    # all_skeleton += np.roll(skeleton, -30, axis=0)

    dist_transform = nd.distance_transform_edt((~skeleton).astype(np.int))

    dist, radprof, weights, unbin_dist, unbin_radprof = \
        radial_profile(model, dist_transform, dist_transform,
                       ((0, 0), (model.shape[0] // 2, model.shape[1] // 2)),
                       img_scale=1.0, auto_cut=True,
                       max_distance=50.0, auto_cut_kwargs={'smooth_size': 3.0,
                                                           'pad_cut': 0})

    npt.assert_equal(dist.max(), 19.25)


def test_radial_profile_autocut_plateau():
    '''
    Test auto-cutting with a plateau and a second fall.
    '''

    model, skeleton = generate_filament_model(shape=160, width=10.0,
                                              amplitude=10.0, background=5.0)

    # Create a second drop-off profile 40 pixels from the center on each side.

    for i, row in enumerate(model[120:].T):

        model[120:, i] = generate_gaussian_profile(np.arange(row.size),
                                                   width=5.0,
                                                   amplitude=5.0,
                                                   background=0.0)

    for i, row in enumerate(model[:40].T):

        model[:40, i] = generate_gaussian_profile(np.arange(row.size),
                                                  width=5.0,
                                                  amplitude=5.0,
                                                  background=0.0)[::-1]

    dist_transform = nd.distance_transform_edt((~skeleton).astype(np.int))

    dist, radprof, weights, unbin_dist, unbin_radprof = \
        radial_profile(model, dist_transform, dist_transform,
                       ((0, 0), (model.shape[0] // 2, model.shape[1] // 2)),
                       img_scale=1.0, auto_cut=True,
                       max_distance=60.0, auto_cut_kwargs={'smooth_size': 3.0,
                                                           'pad_cut': 0,
                                                           'interp_factor': 1})

    # By-eye, this should be 18-19
    npt.assert_almost_equal(dist.max(), 38.201, decimal=3)




import numpy as np
from scipy import ndimage as nd
from astropy.io import fits
import astropy.units as u

from ..width_profiles.profile import _line_profile_coordinates


def generate_filament_model(shape=100, pad_size=0,
                            max_dist=100,
                            theta=0.0, width=10.0,
                            amplitude=1.0, background=0.0,
                            return_hdu=False,
                            phys_pixscale=0.01 * u.pc,
                            phys_beamsize=0.03 * u.pc,
                            distance=250 * u.pc):

    if theta >= np.pi / 2. or theta <= - np.pi / 2.:
        raise ValueError("theta must be between -pi/2 and pi/2")

    yy, xx = np.mgrid[-shape // 2:shape // 2,
                      -shape // 2:shape // 2]

    centers = np.zeros(yy.shape, dtype=bool)
    centers[np.abs(yy - np.tan(theta) * xx) < 1.0] = True

    # Keep positions within the max distance
    cent_point = np.ones(yy.shape, dtype=bool)
    cent_point[yy.shape[0] // 2 + 1, yy.shape[1] // 2 + 1] = False
    mid_dist = nd.distance_transform_edt(cent_point)

    centers[mid_dist > max_dist] = False

    if pad_size > 0:
        centers = np.pad(centers, [(pad_size,) * 2, (pad_size,) * 2],
                         mode='constant', constant_values=0)

    radii = nd.distance_transform_edt(~centers)

    filament = amplitude * np.exp(- (radii ** 2) / (2 * width ** 2)) + \
        background

    if not return_hdu:
        return filament, centers

    pixel_scale = (phys_pixscale / distance).value * u.rad
    beam_size = (phys_beamsize / distance).value * u.rad

    header = create_image_header(pixel_scale, beam_size, filament.shape,
                                 1.e11 * u.Hz, u.K)

    return fits.PrimaryHDU(filament, header), centers


# def make_skeleton_mask(shape=(256, 256), nfils=1):
#     '''
#     Generate a skeleton mask defining positions of filaments.
#     '''


#     mask = np.zeros(shape, dtype=bool)

#     for _ in range(nfils):

#         # Sample random start and end points
#         start = (np.random.randint(shape[0]))

def create_image_header(pixel_scale, beamfwhm, imshape,
                        restfreq, bunit):
    '''
    Create a basic FITS header for an image.

    Adapted from: https://github.com/radio-astro-tools/uvcombine/blob/master/uvcombine/tests/utils.py

    Parameters
    ----------
    pixel_scale : `~astropy.units.Quantity`
        Angular scale of one pixel
    beamfwhm : `~astropy.units.Quantity`
        Angular size for a circular Gaussian beam.
    imshape : tuple
        Shape of the data array.
    restfreq : `~astropy.units.Quantity`
        Rest frequency of the spectral line.
    bunit : `~astropy.units.Unit`
        Unit of intensity.

    Returns
    -------
    header : fits.Header
        FITS Header.
    '''

    header = {'CDELT1': -(pixel_scale).to(u.deg).value,
              'CDELT2': (pixel_scale).to(u.deg).value,
              'BMAJ': beamfwhm.to(u.deg).value,
              'BMIN': beamfwhm.to(u.deg).value,
              'BPA': 0.0,
              'CRPIX1': imshape[0] / 2.,
              'CRPIX2': imshape[1] / 2.,
              'CRVAL1': 0.0,
              'CRVAL2': 0.0,
              'CTYPE1': 'GLON-CAR',
              'CTYPE2': 'GLAT-CAR',
              'CUNIT1': 'deg',
              'CUNIT2': 'deg',
              'CRPIX3': 1,
              'RESTFRQ': restfreq.to(u.Hz).value,
              'BUNIT': bunit.to_string(),
              }

    return fits.Header(header)



# Licensed under an MIT open source license - see LICENSE

import numpy as np
from scipy.stats import scoreatpercentile
import matplotlib.pyplot as p


def rht(mask, radius, ntheta=180, background_percentile=25, verbose=False):
    '''

    Parameters
    ----------

    mask : numpy.ndarray
           Boolean or integer array. Transform performed at all
           non-zero points.

    radius : int
             Radius of circle used around each pixel.

    ntheta : int, optional
             Number of angles to use in transform.

    background_percentile : float, optional
                            Percentile of data to subtract off. Background is
                            due to limits on pixel resolution.

    verbose : bool, optional
        Enables plotting.

    Returns
    -------

    theta : numpy.ndarray
            Angles transform was performed at.

    R : numpy.ndarray
        Transform output.

    quantiles : numpy.ndarray.
        Contains the 25%, mean, and 75% percentile angles.

    '''

    pad_mask = np.pad(mask.astype(float), radius, padwithnans)

    # The theta=0 case isn't handled properly
    theta = np.linspace(np.pi/2., 1.5*np.pi, ntheta)

    # Create a cube of all angle positions
    circle, mesh = circular_region(radius)
    circles_cube = np.empty((ntheta, circle.shape[0], circle.shape[1]))
    for posn, ang in enumerate(theta):
        diff = mesh[0]*np.sin(ang) - mesh[1]*np.cos(ang)
        diff[np.where(np.abs(diff) < 1.0)] = 0
        circles_cube[posn, :, :] = diff

    R = np.zeros((ntheta,))
    x, y = np.where(mask != 0.0)
    for i, j in zip(x, y):
        region = np.tile(circle * pad_mask[i:i+2*radius+1,
                                           j:j+2*radius+1], (ntheta, 1, 1))
        line = region * np.isclose(circles_cube, 0.0)

        if not np.isnan(line).all():
            R = R + np.nansum(np.nansum(line, axis=2), axis=1)

    # Check that the ends are close.
    if np.isclose(R[0], R[-1], rtol=1.0):
        R = R[:-1]
        theta = theta[:-1]
    else:
        raise ValueError("R(-pi/2) should equal R(pi/2). Check input.")

    # You're likely to get a somewhat constant background, so subtract it out
    R = R - np.median(R[R <= scoreatpercentile(R, background_percentile)])
    if (R < 0.0).any():
        R[R < 0.0] = 0.0  # Ignore negative values after subtraction

    # Return to [-pi/2, pi/2] interval and position to the correct zero point.
    theta -= np.pi
    R = np.fliplr(R[:, np.newaxis])

    mean_circ = circ_mean(theta, weights=R)
    twofive, sevenfive = circ_CI(theta, weights=R, u_ci=0.67)
    twofive = twofive[0]
    sevenfive = sevenfive[0]
    quantiles = (twofive, mean_circ, sevenfive)

    if verbose:
        p.subplot(1, 2, 1, polar=True)
        p.plot(2*theta, R, "rD")
        p.plot([2*mean_circ]*2, [0, R.max()], "k")
        p.plot([2*twofive]*2, [0, R.max()], "r")
        p.plot([2*sevenfive]*2, [0, R.max()], "r")
        p.subplot(1, 2, 2)
        p.imshow(mask, cmap="binary", origin='lower')
        p.show()

    return theta, R, quantiles


def circular_region(radius):
    '''
    Create a circle of a given radius.
    Values are NaNs outside of the circle.

    Parameters
    ----------
    radius : int
        Circle radius.

    Returns
    -------
    circle : numpy.ndarray
        Array containing the circle.
    [xx, yy] : numpy.ndarray
        Grids used to create the circle.
    '''
    xx, yy = np.mgrid[-radius:radius+1, -radius:radius+1]

    circle = xx**2. + yy**2.
    circle = circle < radius**2.

    circle = circle.astype(float)
    circle[np.where(circle == 0.)] = np.NaN

    return circle, [xx, yy]


def padwithnans(vector, pad_width, iaxis, kwargs):
    vector[:pad_width[0]] = np.NaN
    vector[-pad_width[1]:] = np.NaN
    return vector


def find_nearest(array, value):
    idx = (np.abs(array-value)).argmin()
    return array[idx]


def find_nearest_posn(array, value):
    idx = (np.abs(array-value)).argmin()
    return idx


def circ_mean(theta, weights=None):
    """
    Calculates the median of a set of angles on the circle and returns
    a value on the interval from (-pi, pi].  Angles expected in radians.

    Parameters
    ----------
    """

    if len(theta.shape) == 1:
        theta = theta[:, np.newaxis]

    if weights is None:
        weights = np.ones(theta.shape)

    medangle = np.arctan2(np.nansum(np.sin(2*theta) * weights),
                          np.nansum(np.cos(2*theta) * weights))

    medangle /= 2.0

    return medangle


def fourier_shifter(x, shift, axis):
    '''
    Shift an array by some value along its axis.
    '''
    ftx = np.fft.fft(x, axis=axis)
    m = np.fft.fftfreq(x.shape[axis])
    # m_shape = [1] * x.ndim
    # m_shape[axis] = m.shape[0]
    # m = m.reshape(m_shape)
    slices = [slice(None) if ii == axis else None for ii in range(x.ndim)]
    m = m[tuple(slices)]
    phase = np.exp(-2 * np.pi * m * 1j * shift)
    x2 = np.real(np.fft.ifft(ftx * phase, axis=axis))
    return x2


def circ_CI(theta, weights=None, u_ci=0.67, axis=0):
    '''

    '''

    if len(theta.shape) == 1:
        theta = theta[:, np.newaxis]

    if weights is None:
        weights = np.ones(theta.shape)
    else:
        if len(weights.shape) == 1:
            weights = weights[:, np.newaxis]

    assert theta.shape == weights.shape

    # Normalize weights
    weights /= np.sum(weights, axis=axis)

    mean_ang = circ_mean(theta, weights=weights)

    # Now center the data around the mean to find the CI intervals
    diff_val = np.diff(theta[:2, 0])[0]

    theta_mid = theta[theta.shape[0] // 2]

    diff_posn = - (theta_mid - mean_ang) / diff_val

    theta_copy = fourier_shifter(theta, diff_posn, axis=0)

    vec_length2 = np.sum(weights * np.cos(theta_copy), axis=axis)**2. + \
        np.sum(weights * np.sin(theta_copy), axis=axis)**2.

    alpha = np.sum(weights * np.cos(2 * theta_copy), axis=axis)

    var_w = (1 - alpha) / (4 * vec_length2)

    # Make sure the CI stays within the interval. Otherwise assign it to
    # pi/2 (largest possible on interval of pi)
    sin_arg = u_ci * np.sqrt(2 * var_w)

    if sin_arg <= 1:
        ci = np.arcsin(sin_arg)
    else:
        ci = np.pi / 2.

    samp_cis = np.vstack([mean_ang - ci, mean_ang + ci])

    return samp_cis



# Licensed under an MIT open source license - see LICENSE

import pytest

import numpy as np
import numpy.testing as npt
import astropy.units as u
from astropy.io.fits import PrimaryHDU

try:
    from spectral_cube import Projection, Slice
    SPECTRALCUBE_INSTALL = True
except ImportError:
    SPECTRALCUBE_INSTALL = False


from ..io_funcs import input_data

from ._testing_data import *


def test_array_input():

    output = input_data(img)

    npt.assert_equal(img, output["data"].value)
    assert output['data'].unit == u.dimensionless_unscaled


def test_array_input_withheader():

    new_hdr = hdr.copy()
    new_hdr['BUNIT'] = 'K'
    output = input_data(img, header=new_hdr)

    npt.assert_equal(img, output["data"].value)
    assert output['data'].unit == u.K
    npt.assert_equal(new_hdr, output["header"])


def test_quantity_input():

    quant = img * u.K

    output = input_data(quant)

    npt.assert_equal(img, output["data"].value)
    assert output['data'].unit == u.K


def test_quantity_input_withheader():

    quant = img * u.K

    # Give the header a different BUNIT. Always use the unit
    # attached to the Quantity object
    new_hdr = hdr.copy()
    new_hdr['BUNIT'] = 'Jy/beam'

    output = input_data(quant, header=new_hdr)

    npt.assert_equal(img, output["data"].value)
    assert output['data'].unit == u.K

    # The header should now have K set
    assert output['header']['BUNIT'] == 'K'


def test_HDU_input():

    hdu = PrimaryHDU(img, header=hdr)

    output = input_data(hdu)

    npt.assert_equal(img, output["data"].value)
    assert output['data'].unit == u.dimensionless_unscaled
    npt.assert_equal(hdr, output["header"])


def test_HDU_input_withbunit():

    hdr['BUNIT'] = 'K'
    hdu = PrimaryHDU(img, header=hdr)

    output = input_data(hdu)

    npt.assert_equal(img, output["data"].value)
    assert output['data'].unit == u.K
    npt.assert_equal(hdr, output["header"])


@pytest.mark.skipif("not SPECTRALCUBE_INSTALL")
def test_SC_inputs():

    hdr['BUNIT'] = 'K'
    hdu = PrimaryHDU(img, header=hdr)

    proj = Projection.from_hdu(hdu)

    output = input_data(proj)

    npt.assert_equal(img, output["data"].value)
    assert output['data'].unit == u.K
    npt.assert_equal(proj.header, output["header"])

    slic = Slice.from_hdu(hdu)

    output = input_data(slic)

    npt.assert_equal(img, output["data"].value)
    assert output['data'].unit == u.K
    npt.assert_equal(slic.header, output["header"])


def test_3D_input():

    with pytest.raises(TypeError):
        input_data(np.ones((3, ) * 3))


def test_3D_squeezable_input():

    output = input_data(np.ones((3, 3, 1)))

    npt.assert_equal(np.ones((3, 3)), output["data"].value)
    assert output['data'].unit == u.dimensionless_unscaled



# Licensed under an MIT open source license - see LICENSE

from .length import *
from .utilities import distance

import numpy as np
import scipy.ndimage as nd
import matplotlib.pyplot as p
import copy


def isolateregions(binary_array, size_threshold=0, pad_size=0,
                   fill_hole=False, rel_size=0.1, morph_smooth=False):
    '''

    Labels regions in a boolean array and returns individual arrays for each
    region. Regions below a threshold can optionally be removed. Small holes
    may also be filled in.

    Parameters
    ----------
    binary_array : numpy.ndarray
        A binary array of regions.
    size_threshold : int, optional
        Sets the pixel size on the size of regions.
    pad_size : int, optional
        Padding to be added to the individual arrays.
    fill_hole : int, optional
        Enables hole filling.
    rel_size : float or int, optional
        If < 1.0, sets the minimum size a hole must be relative to the area
        of the mask. Otherwise, this is the maximum number of pixels the hole
        must have to be deleted.
    morph_smooth : bool, optional
        Morphologically smooth the image using a binar opening and closing.

    Returns
    -------
    output_arrays : list
        Regions separated into individual arrays.
    num : int
        Number of filaments
    corners : list
        Contains the indices where each skeleton array was taken from
        the original.

    '''

    output_arrays = []
    corners = []

    # Label skeletons
    labels, num = nd.label(binary_array, eight_con())

    # Remove skeletons which have fewer pixels than the threshold.
    if size_threshold != 0:
        sums = nd.sum(binary_array, labels, range(1, num + 1))
        remove_fils = np.where(sums <= size_threshold)[0]
        for lab in remove_fils:
            binary_array[np.where(labels == lab + 1)] = 0

        # Relabel after deleting short skeletons.
        labels, num = nd.label(binary_array, eight_con())

    # Split each skeleton into its own array.
    for n in range(1, num + 1):
        x, y = np.where(labels == n)
        # Make an array shaped to the skeletons size and padded on each edge
        # the +1 is because, e.g., range(0, 5) only has 5 elements, but the
        # indices we're using are range(0, 6)
        shapes = (x.max() - x.min() + 2 * pad_size + 1,
                  y.max() - y.min() + 2 * pad_size + 1)
        eachfil = np.zeros(shapes)
        eachfil[x - x.min() + pad_size, y - y.min() + pad_size] = 1
        # Fill in small holes
        if fill_hole:
            eachfil = _fix_small_holes(eachfil, rel_size=rel_size)
        if morph_smooth:
            eachfil = nd.binary_opening(eachfil, np.ones((3, 3)))
            eachfil = nd.binary_closing(eachfil, np.ones((3, 3)))
        output_arrays.append(eachfil)
        # Keep the coordinates from the original image
        lower = (max(0, x.min() - pad_size), max(0, y.min() - pad_size))
        upper = (x.max() + pad_size + 1, y.max() + pad_size + 1)
        corners.append([lower, upper])

    return output_arrays, num, corners


def find_filpix(branches, labelfil, final=True, debug=False):
    '''

    Identifies the types of pixels in the given skeletons. Identification is
    based on the connectivity of the pixel.

    Parameters
    ----------
    branches : list
        Contains the number of branches in each skeleton.
    labelfil : list
        Contains the arrays of each skeleton.
    final : bool, optional
        If true, corner points, intersections, and body points are all
        labeled as a body point for use when the skeletons have already
        been cleaned.
    debug : bool, optional
        Enable to print out (a lot) of extra info on pixel classification.

    Returns
    -------
    fila_pts : list
        All points on the body of each skeleton.
    inters : list
        All points associated with an intersection in each skeleton.
    labelfil : list
       Contains the arrays of each skeleton where all intersections
       have been removed.
    endpts_return : list
        The end points of each branch of each skeleton.
  '''

    initslices = []
    initlist = []
    shiftlist = []
    sublist = []
    endpts = []
    blockpts = []
    bodypts = []
    slices = []
    vallist = []
    shiftvallist = []
    cornerpts = []
    subvallist = []
    subslist = []
    pix = []
    filpix = []
    intertemps = []
    fila_pts = []
    inters = []
    repeat = []
    temp_group = []
    all_pts = []
    pairs = []
    endpts_return = []

    for k in range(1, branches + 1):
        x, y = np.where(labelfil == k)
        for i in range(len(x)):
            if x[i] < labelfil.shape[0] - 1 and y[i] < labelfil.shape[1] - 1:
                pix.append((x[i], y[i]))
                initslices.append(np.array([[labelfil[x[i] - 1, y[i] + 1],
                                             labelfil[x[i], y[i] + 1],
                                             labelfil[x[i] + 1, y[i] + 1]],
                                            [labelfil[x[i] - 1, y[i]], 0,
                                                labelfil[x[i] + 1, y[i]]],
                                            [labelfil[x[i] - 1, y[i] - 1],
                                             labelfil[x[i], y[i] - 1],
                                             labelfil[x[i] + 1, y[i] - 1]]]))

        filpix.append(pix)
        slices.append(initslices)
        initslices = []
        pix = []

    for i in range(len(slices)):
        for k in range(len(slices[i])):
            initlist.append([slices[i][k][0, 0],
                             slices[i][k][0, 1],
                             slices[i][k][0, 2],
                             slices[i][k][1, 2],
                             slices[i][k][2, 2],
                             slices[i][k][2, 1],
                             slices[i][k][2, 0],
                             slices[i][k][1, 0]])
        vallist.append(initlist)
        initlist = []

    for i in range(len(slices)):
        for k in range(len(slices[i])):
            shiftlist.append(shifter(vallist[i][k], 1))
        shiftvallist.append(shiftlist)
        shiftlist = []

    for k in range(len(slices)):
        for i in range(len(vallist[k])):
            for j in range(8):
                sublist.append(
                    int(vallist[k][i][j]) - int(shiftvallist[k][i][j]))
            subslist.append(sublist)
            sublist = []
        subvallist.append(subslist)
        subslist = []

    # x represents the subtracted list (step-ups) and y is the values of the
    # surrounding pixels. The categories of pixels are ENDPTS (x<=1),
    # BODYPTS (x=2,y=2),CORNERPTS (x=2,y=3),BLOCKPTS (x=3,y>=4), and
    # INTERPTS (x>=3).
    # A cornerpt is [*,0,0] (*s) associated with an intersection,
    # but their exclusion from
    #   [1,*,0] the intersection keeps eight-connectivity, they are included
    #   [0,1,0] intersections for this reason.
    # A blockpt is  [1,0,1] They are typically found in a group of four,
    # where all four
    #   [0,*,*] constitute a single intersection.
    #   [1,*,*]
    # A T-pt has the same connectivity as a block point, but with two 8-conns
    # [*, *, *]
    # [0, 1, 0]
    # The "final" designation is used when finding the final branch lengths.
    # At this point, blockpts and cornerpts should be eliminated.
    for k in range(branches):
        for l in range(len(filpix[k])):
            x = [j for j, y in enumerate(subvallist[k][l]) if y == k + 1]
            y = [j for j, z in enumerate(vallist[k][l]) if z == k + 1]

            if len(x) <= 1:
                if debug:
                    print("End pt. {}".format(filpix[k][l]))
                endpts.append(filpix[k][l])
                endpts_return.append(filpix[k][l])
            elif len(x) == 2:
                if final:
                    bodypts.append(filpix[k][l])
                else:
                    if len(y) == 2:
                        if debug:
                            print("Body pt. {}".format(filpix[k][l]))
                        bodypts.append(filpix[k][l])

                    elif is_tpoint(vallist[k][l]):
                        # If there are only 3 connections to the t-point, it
                        # is an end point
                        if len(y) == 3:
                            if debug:
                                print("T-point end {}".format(filpix[k][l]))
                            endpts.append(filpix[k][l])
                            endpts_return.append(filpix[k][l])
                        # If there are 4, it is a body point
                        elif len(y) == 4:
                            if debug:
                                print("T-point body {}".format(filpix[k][l]))

                            bodypts.append(filpix[k][l])
                        # Otherwise it is a part of an intersection
                        else:
                            if debug:
                                print("T-point inter {}".format(filpix[k][l]))
                            intertemps.append(filpix[k][l])
                    elif is_blockpoint(vallist[k][l]):
                        if debug:
                            print("Block pt. {}".format(filpix[k][l]))
                        blockpts.append(filpix[k][l])
                    else:
                        if debug:
                            print("Corner pt. {}".format(filpix[k][l]))
                        cornerpts.append(filpix[k][l])
            elif len(x) >= 3:
                if debug:
                    print("Inter pt. {}".format(filpix[k][l]))
                intertemps.append(filpix[k][l])
        endpts = list(set(endpts))
        bodypts = list(set(bodypts))
        dups = set(endpts) & set(bodypts)
        if len(dups) > 0:
            for i in dups:
                bodypts.remove(i)
        # Cornerpts without a partner diagonally attached can be included as a
        # bodypt.
        if debug:
            print("Cornerpts: {}".format(cornerpts))

        if len(cornerpts) > 0:
            deleted_cornerpts = []

            for i, j in zip(cornerpts, cornerpts):
                if i != j:
                    if distance(i[0], j[0], i[1], j[1]) == np.sqrt(2.0):
                        proximity = [(i[0], i[1] - 1),
                                     (i[0], i[1] + 1),
                                     (i[0] - 1, i[1]),
                                     (i[0] + 1, i[1]),
                                     (i[0] - 1, i[1] + 1),
                                     (i[0] + 1, i[1] + 1),
                                     (i[0] - 1, i[1] - 1),
                                     (i[0] + 1, i[1] - 1)]
                        match = set(intertemps) & set(proximity)
                        if len(match) == 1:
                            print("MATCH")
                            bodypts.extend([i, j])
                            # pairs.append([i, j])
                            deleted_cornerpts.append(i)
                            deleted_cornerpts.append(j)
            cornerpts = list(set(cornerpts).difference(set(deleted_cornerpts)))

        if len(cornerpts) > 0:
            for l in cornerpts:
                proximity = [(l[0], l[1] - 1),
                             (l[0], l[1] + 1),
                             (l[0] - 1, l[1]),
                             (l[0] + 1, l[1]),
                             (l[0] - 1, l[1] + 1),
                             (l[0] + 1, l[1] + 1),
                             (l[0] - 1, l[1] - 1),
                             (l[0] + 1, l[1] - 1)]

                # Check if the matching corner point is an end point
                # Otherwise the pixel will be combined into a 2-pixel intersec
                match_ends = set(endpts) & set(proximity[-4:])
                if len(match_ends) == 1:
                    fila_pts.append(endpts + bodypts + [l])
                    continue

                match = set(intertemps) & set(proximity)
                if len(match) == 1:
                    intertemps.append(l)
                    fila_pts.append(endpts + bodypts)
                else:
                    fila_pts.append(endpts + bodypts + [l])
        else:
            fila_pts.append(endpts + bodypts)

        # Reset lists
        cornerpts = []
        endpts = []
        bodypts = []

        if len(pairs) > 0:
            for i in range(len(pairs)):
                for j in pairs[i]:
                    all_pts.append(j)
        if len(blockpts) > 0:
            for i in blockpts:
                all_pts.append(i)
        if len(intertemps) > 0:
            for i in intertemps:
                all_pts.append(i)
        # Pairs of cornerpts, blockpts, and interpts are combined into an
        # array. If there is eight connectivity between them, they are labelled
        # as a single intersection.
        arr = np.zeros((labelfil.shape))
        for z in all_pts:
            labelfil[z[0], z[1]] = 0
            arr[z[0], z[1]] = 1
        lab, nums = nd.label(arr, eight_con())
        for k in range(1, nums + 1):
            objs_pix = np.where(lab == k)
            for l in range(len(objs_pix[0])):
                temp_group.append((objs_pix[0][l], objs_pix[1][l]))
            inters.append(temp_group)
            temp_group = []
    for i in range(len(inters) - 1):
        if inters[i] == inters[i + 1]:
            repeat.append(inters[i])
    for i in repeat:
        inters.remove(i)

    if debug:
        print("Fila pts: {}".format(fila_pts))
        print("Intersections: {}".format(inters))
        print("End pts: {}".format(endpts_return))
        print(labelfil)

    return fila_pts, inters, labelfil, endpts_return


def find_extran(branches, labelfil, debug=False):
    '''
    Identify pixels that are not necessary to keep the connectivity of the
    skeleton. It uses the same labeling process as find_filpix. Extraneous
    pixels tend to be those from former intersections, whose attached branch
    was eliminated in the cleaning process.

    Parameters
    ----------
    branches : list
        Contains the number of branches in each skeleton.
    labelfil : list
        Contains arrays of the labeled versions of each skeleton.
    debug : bool, optional
        Enable plotting of each filament array to visualize where the deleted
        pixels are.

    Returns
    -------
    labelfil : list
       Contains the updated labeled arrays with extraneous pieces
       removed.
    '''

    initslices = []
    initlist = []
    shiftlist = []
    sublist = []
    extran = []
    slices = []
    vallist = []
    shiftvallist = []
    subvallist = []
    subslist = []
    pix = []
    filpix = []

    for k in range(1, branches + 1):
        x, y = np.where(labelfil == k)
        for i in range(len(x)):
            if x[i] < labelfil.shape[0] - 1 and y[i] < labelfil.shape[1] - 1:
                pix.append((x[i], y[i]))
                initslices.append(np.array([[labelfil[x[i] - 1, y[i] + 1],
                                             labelfil[x[i], y[i] + 1],
                                             labelfil[x[i] + 1, y[i] + 1]],
                                            [labelfil[x[i] - 1, y[i]], 0,
                                             labelfil[x[i] + 1, y[i]]],
                                            [labelfil[x[i] - 1, y[i] - 1],
                                             labelfil[x[i], y[i] - 1],
                                             labelfil[x[i] + 1, y[i] - 1]]]))

        filpix.append(pix)
        slices.append(initslices)
        initslices = []
        pix = []

    for i in range(len(slices)):
        for k in range(len(slices[i])):
            initlist.append([slices[i][k][0, 0],
                             slices[i][k][0, 1],
                             slices[i][k][0, 2],
                             slices[i][k][1, 2],
                             slices[i][k][2, 2],
                             slices[i][k][2, 1],
                             slices[i][k][2, 0],
                             slices[i][k][1, 0]])
        vallist.append(initlist)
        initlist = []

    for i in range(len(slices)):
        for k in range(len(slices[i])):
            shiftlist.append(shifter(vallist[i][k], 1))
        shiftvallist.append(shiftlist)
        shiftlist = []

    for k in range(len(slices)):
        for i in range(len(vallist[k])):
            for j in range(8):
                sublist.append(
                    int(vallist[k][i][j]) - int(shiftvallist[k][i][j]))
            subslist.append(sublist)
            sublist = []
        subvallist.append(subslist)
        subslist = []

    for k in range(len(slices)):
        for l in range(len(filpix[k])):
            x = [j for j, y in enumerate(subvallist[k][l]) if y == k + 1]
            y = [j for j, z in enumerate(vallist[k][l]) if z == k + 1]

            if len(x) == 0:
                if debug:
                    print("Extran removal unconnect: {}".format(filpix[k][l]))
                labelfil[filpix[k][l][0], filpix[k][l][1]] = 0
                extran.append(filpix[k][l])

            if len(x) == 1:
                if len(y) >= 2:
                    if debug:
                        print("Extran removal: {}".format(filpix[k][l]))
                    extran.append(filpix[k][l])
                    labelfil[filpix[k][l][0], filpix[k][l][1]] = 0
        # if len(extran) >= 2:
        #     for i in extran:
        #         for j in extran:
        #             if i != j:
        #                 if distance(i[0], j[0], i[1], j[1]) == np.sqrt(2.0):
        #                     proximity = [(i[0], i[1] - 1),
        #                                  (i[0], i[1] + 1),
        #                                  (i[0] - 1, i[1]),
        #                                  (i[0] + 1, i[1]),
        #                                  (i[0] - 1, i[1] + 1),
        #                                  (i[0] + 1, i[1] + 1),
        #                                  (i[0] - 1, i[1] - 1),
        #                                  (i[0] + 1, i[1] - 1)]
        #                     match = set(filpix[k]) & set(proximity)
        #                     if len(match) > 0:
        #                         for z in match:
        #                             labelfil[z[0], z[1]] = 0

    if debug:
        import matplotlib.pyplot as plt
        plt.imshow(labelfil, origin='lower')
        for pix in extran:
            plt.plot(pix[1], pix[0], 'bD')
        plt.draw()
        raw_input("?")
        plt.clf()

    return labelfil


######################################################################
# Wrapper Functions
######################################################################


def pix_identify(isolatefilarr, num, debug=False):
    '''
    This function is essentially a wrapper on find_filpix. It returns the
    outputs of find_filpix in the form that are used during the analysis.

    Parameters
    ----------
    isolatefilarr : list
        Contains individual arrays of each skeleton.
    num  : int
        The number of skeletons.
    debug : bool, optional
        Print out identification steps in find_filpix.

    Returns
    -------
    interpts : list
        Contains lists of all intersections points in each skeleton.
    hubs : list
        Contains the number of intersections in each filament. This is
        useful for identifying those with no intersections as their analysis
        is straight-forward.
    ends : list
        Contains the positions of all end points in each skeleton.
    filbranches : list
        Contains the number of branches in each skeleton.
    labelisofil : list
        Contains individual arrays for each skeleton where the
        branches are labeled and the intersections have been removed.
    '''

    interpts = []
    hubs = []
    ends = []
    filbranches = []
    labelisofil = []

    for n in range(num):
        funcreturn = find_filpix(1, isolatefilarr[n], final=False, debug=debug)
        interpts.append(funcreturn[1])
        hubs.append(len(funcreturn[1]))
        isolatefilarr.pop(n)
        isolatefilarr.insert(n, funcreturn[2])
        ends.append(funcreturn[3])

        # isolatefilarr contains end and body pts. We need to make sure the end
        # points don't touch, and if they do, label them independently of the
        # others. Touching endpoints can only occur for 1-pixel branches

        # First label the end points
        ends_arr = np.zeros_like(isolatefilarr[n])
        for end in ends[n]:
            ends_arr[end[0], end[1]] = True
        end_label, num_end = nd.label(ends_arr, eight_con())
        # If none are connected, label normally
        if len(ends[n]) == num_end:
            label_branch, num_branch = nd.label(isolatefilarr[n], eight_con())
        else:
            # Find the connected ends, then label and remove them
            conn_ends = np.where(nd.sum(ends_arr, end_label,
                                        range(1, num_end + 1)) > 1)[0] + 1
            if conn_ends.size == 0:
                raise ValueError("This should not be possible, since "
                                 "num_end != len(ends[n]), but no connected"
                                 " structure was found? Possible bug.")

            fil_arr = isolatefilarr[n].copy()

            end_branches = []
            for conn in conn_ends:
                end_posns = np.where(end_label == conn)
                for posn in zip(*end_posns):
                    end_branches.append(posn)
                fil_arr[end_posns] = False

            # Label the version without those ends
            label_branch, num_branch = nd.label(fil_arr, eight_con())

            # Now re-add those connected ends with a new label
            for i, posn in enumerate(end_branches):
                label_branch[posn[0], posn[1]] = num_branch + i + 1

            num_branch = label_branch.max()

        filbranches.append(num_branch)
        labelisofil.append(label_branch)

    return interpts, hubs, ends, filbranches, labelisofil


def extremum_pts(labelisofil, extremum, ends):
    '''
    This function returns the the farthest extents of each filament. This
    is useful for determining how well the shortest path algorithm has worked.

    Parameters
    ----------
    labelisofil : list
        Contains individual arrays for each skeleton.
    extremum : list
       Contains the extents as determined by the shortest
       path algorithm.
    ends : list
        Contains the positions of each end point in eahch filament.

    Returns
    -------
    extrem_pts : list
        Contains the indices of the extremum points.
    '''

    num = len(labelisofil)
    extrem_pts = []

    for n in range(num):
        per_fil = []
        for i, j in ends[n]:
            if labelisofil[n][i, j] == extremum[n][0] or labelisofil[n][i, j] == extremum[n][1]:
                per_fil.append([i, j])
        extrem_pts.append(per_fil)

    return extrem_pts


def make_final_skeletons(labelisofil, inters, verbose=False, save_png=False,
                         save_name=None):
    '''
    Creates the final skeletons outputted by the algorithm.

    Parameters
    ----------
    labelisofil : list
        List of labeled skeletons.
    inters : list
        Positions of the intersections in each skeleton.
    verbose : bool, optional
        Enables plotting of the final skeleton.
    save_png : bool, optional
        Saves the plot made in verbose mode. Disabled by default.
    save_name : str, optional
        For use when ``save_png`` is enabled.
        **MUST be specified when ``save_png`` is enabled.**

    Returns
    -------
    filament_arrays : list
        List of the final skeletons.
    '''

    filament_arrays = []

    for n, (skel_array, intersec) in enumerate(zip(labelisofil, inters)):
        copy_array = np.zeros(skel_array.shape, dtype=int)

        for inter in intersec:
            for pts in inter:
                x, y = pts
                copy_array[x, y] = 1

        copy_array[np.where(skel_array >= 1)] = 1

        cleaned_array = find_extran(1, copy_array)

        filament_arrays.append(cleaned_array)

        if verbose or save_png:
            if save_png and save_name is None:
                ValueError("Must give a save_name when save_png is enabled. No"
                           " plots will be created.")

            p.clf()
            p.imshow(cleaned_array, origin='lower', interpolation='nearest')

            if save_png:
                p.savefig(save_name)
                p.close()
            if verbose:
                p.show()
            if in_ipynb():
                p.clf()

    return filament_arrays


def recombine_skeletons(skeletons, offsets, orig_size, pad_size):
    '''
    Takes a list of skeleton arrays and combines them back into
    the original array.

    Parameters
    ----------
    skeletons : list
        Arrays of each skeleton.
    offsets : list
        Coordinates where the skeleton arrays have been sliced from the
        image.
    orig_size : tuple
        Size of the image.
    pad_size : int
        Size of the array padding.

    Returns
    -------
    master_array : numpy.ndarray
        Contains all skeletons placed in their original positions in the image.
    '''

    num = len(skeletons)

    master_array = np.zeros(orig_size)
    for n in range(num):
        # These are the coordinates of the bottom left in the master array.
        x_off, y_off = offsets[n][0]
        x_top, y_top = offsets[n][1]

        # Now check if padding will put the array outside of the original array
        # size
        excess_x_top = x_top - orig_size[0]

        excess_y_top = y_top - orig_size[1]

        copy_skeleton = copy.copy(skeletons[n])

        if excess_x_top > 0:
            copy_skeleton = copy_skeleton[:-excess_x_top, :]

        if excess_y_top > 0:
            copy_skeleton = copy_skeleton[:, :-excess_y_top]

        if x_off < 0:
            copy_skeleton = copy_skeleton[-x_off:, :]
            x_off = 0

        if y_off < 0:
            copy_skeleton = copy_skeleton[:, -y_off:]
            y_off = 0

        x, y = np.where(copy_skeleton >= 1)
        for i in range(len(x)):
            master_array[x[i] + x_off, y[i] + y_off] = 1

    return master_array


def _fix_small_holes(mask_array, rel_size=0.1):
    '''
    Helper function to remove only small holes within a masked region.

    Parameters
    ----------
    mask_array : numpy.ndarray
        Array containing the masked region.

    rel_size : float, optional
        If < 1.0, sets the minimum size a hole must be relative to the area
        of the mask. Otherwise, this is the maximum number of pixels the hole
        must have to be deleted.

    Returns
    -------
    mask_array : numpy.ndarray
        Altered array.
    '''

    if rel_size <= 0.0:
        raise ValueError("rel_size must be positive.")
    elif rel_size > 1.0:
        pixel_flag = True
    else:
        pixel_flag = False

    # Find the region area
    reg_area = len(np.where(mask_array == 1)[0])

    # Label the holes
    holes = np.logical_not(mask_array).astype(float)
    lab_holes, n_holes = nd.label(holes, eight_con())

    # If no holes, return
    if n_holes == 1:
        return mask_array

    # Ignore area outside of the region.
    out_label = lab_holes[0, 0]
    # Set size to be just larger than the region. Thus it can never be
    # deleted.
    holes[np.where(lab_holes == out_label)] = reg_area + 1.

    # Sum up the regions and find holes smaller than the threshold.
    sums = nd.sum(holes, lab_holes, range(1, n_holes + 1))
    if pixel_flag:  # Use number of pixels
        delete_holes = np.where(sums < rel_size)[0]
    else:  # Use relative size of holes.
        delete_holes = np.where(sums / reg_area < rel_size)[0]

    # Return if there is nothing to delete.
    if len(delete_holes) == 0:
        return mask_array

    # Add one to take into account 0 in list if object label 1.
    delete_holes += 1
    for label in delete_holes:
        mask_array[np.where(lab_holes == label)] = 1

    return mask_array


def is_blockpoint(vallist):
    '''
    Determine if point is part of a block:
    [X X]
    [X X]

    Will have 3 connected sides, with one as an 8-connection.
    '''

    vals = np.array(vallist)

    if vals.sum() < 3:
        return False

    arrangements = [np.array([0, 1, 7]), np.array([1, 2, 3]),
                    np.array([3, 4, 5]), np.array([5, 6, 7])]

    posns = np.where(vals)[0]

    # Check if all 3 in an arrangement are within the vallist
    for arrange in arrangements:
        if np.in1d(posns, arrange).sum() == 3:
            return True

    return False


def is_tpoint(vallist):
    '''
    Determine if point is part of a block:
    [X X X]
    [0 X 0]
    And all 90 deg rotation of this shape

    If there are only 3 connections, this is an end point. If there are 4,
    it is a body point, and if there are 5, it remains an intersection

    '''

    vals = np.array(vallist)

    if vals.sum() < 3:
        return False

    arrangements = [np.array([0, 6, 7]), np.array([0, 1, 2]),
                    np.array([2, 3, 4]), np.array([4, 5, 6])]

    posns = np.where(vals)[0]

    # Check if all 3 in an arrangement are within the vallist
    for arrange in arrangements:
        if np.in1d(posns, arrange).sum() == 3:
            return True

    return False


def merge_nodes(node, G):
    '''
    Combine a node into its neighbors.
    '''

    neigb = list(G[node])

    if len(neigb) != 2:
        return G

    new_weight = G[node][neigb[0]]['weight'] + \
        G[node][neigb[1]]['weight']

    G.remove_node(node)
    G.add_edge(neigb[0], neigb[1], weight=new_weight)

    return G




import scipy.ndimage as nd
from scipy.optimize import curve_fit
import numpy as np
import matplotlib.pyplot as p
import astropy.units as u
from astropy.table import QTable

from .profile import profile_line

eight_conn = np.ones((3, 3))

end_structs = [np.array([[1, 0, 0],
                         [0, 1, 0],
                         [0, 0, 0]]),
               np.array([[0, 1, 0],
                         [0, 1, 0],
                         [0, 0, 0]]),
               np.array([[0, 0, 1],
                         [0, 1, 0],
                         [0, 0, 0]]),
               np.array([[0, 0, 0],
                         [1, 1, 0],
                         [0, 0, 0]]),
               np.array([[0, 0, 0],
                         [0, 1, 1],
                         [0, 0, 0]]),
               np.array([[0, 0, 0],
                         [0, 1, 0],
                         [1, 0, 0]]),
               np.array([[0, 0, 0],
                         [0, 1, 0],
                         [0, 1, 0]]),
               np.array([[0, 0, 0],
                         [0, 1, 0],
                         [0, 0, 1]])]

four_conn_posns = [1, 3, 5, 7]
eight_conn_posns = [0, 2, 6, 8]


def filament_profile(skeleton, image, pixscale, max_dist=0.025 * u.pc,
                     distance=250. * u.pc, num_avg=3, verbose=False,
                     bright_unit="Jy km/s", noise=None, fit_profiles=True):
    '''
    Calculate radial profiles along the main extent of a skeleton (ie. the
    longest path). The skeleton must contain a single branch with no
    intersections.

    Parameters
    ----------
    skeleton : np.ndarray
        Boolean array containing the skeleton
    image : np.ndarray
        Image to compute the profiles from. Must match the spatial extent
        of the skeleton array.
    pixscale : `~astropy.units.Quantity`
        Angular size of a pixel in the image. Must have units equivalent to
        degrees.
    max_dist : astropy Quantity, optional
        The angular or physical (when distance is given) extent to create the
        profile away from the centre skeleton pixel. The entire profile will
        be twice this value (for each side of the profile).
    distance : astropy Quantity, optional
        Physical distance to the region in the image. If None is given,
        results will be in angular units based on the header.
    num_avg : int, optional
        Number of points before and after a pixel that is used when computing
        the normal vector. Using at least three points is recommended due to
        small pixel instabilities in the skeletons.
    verbose : bool, optional
        Enable plotting of the profile and the accompanying for each pixel in
        the skeleton.
    bright_unit : string or astropy Unit
        Brightness unit of the image.
    noise : np.ndarray, optional
        RMS array for the accompanying image. When provided, the errors
        are calculated along each of the profiles and used as weights in the
        fitting.
    fit_profiles : bool, optional
        When enabled, fits a Gaussian model to the profiles. Otherwise only
        the profiles are returned.

    Returns
    -------
    line_distances : list
        Distances along the profiles.
    line_profiles : list
        Radial profiles.
    profile_extents : list
        Contains the pixel position of the start of the profile,
        the skeleton pixel, and the end of the profile.
    tab : astropy QTable
        Table of the fit results and errors with appropriate units.
    '''

    deg_per_pix = pixscale.to(u.deg) / u.pixel

    if distance is not None:
        phys_per_pix = distance * (np.pi / 180.) * deg_per_pix / u.deg

        max_pixel = (max_dist / phys_per_pix).value

    else:
        # max_dist should then be in pixel or angular units
        if not isinstance(max_dist, u.Quantity):
            # Assume pixels
            max_pixel = max_dist
        else:
            try:
                max_pixel = max_dist.to(u.pix).value
            except u.UnitConversionError:
                # In angular units
                equiv = [(u.pixel, u.deg, lambda x: x / (pixscale * u.pix),
                          lambda x: x * pixscale * u.pix)]
                max_pixel = max_dist.to(u.pix, equivalencies=equiv).value

    if bright_unit is None:
        bright_unit = u.dimensionless_unscaled
    elif isinstance(bright_unit, str):
        bright_unit = u.Unit(bright_unit)
    elif isinstance(bright_unit, u.UnitBase):
        pass
    else:
        raise TypeError("bright_unit must be compatible with astropy.units.")

    # Make sure the noise array is the same shape
    if noise is not None:
        assert noise.shape == image.shape

    # Get the points in the skeleton (in order)
    skel_pts = walk_through_skeleton(skeleton)

    line_profiles = []
    line_distances = []
    profile_extents = []
    profile_fits = []
    red_chisqs = []

    for j, i in enumerate(range(num_avg, len(skel_pts) - num_avg)):
        # Calculate the normal direction from the surrounding pixels
        pt1 = avg_pts([skel_pts[i + j] for j in range(-num_avg, 0)])
        pt2 = avg_pts([skel_pts[i + j] for j in range(1, num_avg + 1)])

        vec = np.array([float(x2 - x1) for x2, x1 in
                        zip(pt1, pt2)])
        vec /= np.linalg.norm(vec)

        per_vec = perpendicular(vec)

        line_pts = find_path_ends(skel_pts[i], max_pixel, per_vec)

        left_profile, left_dists = \
            profile_line(image, skel_pts[i], line_pts[0])
        right_profile, right_dists = \
            profile_line(image, skel_pts[i], line_pts[1])

        total_profile = np.append(left_profile[::-1], right_profile) * \
            bright_unit

        if noise is not None:
            left_profile, _ = \
                profile_line(noise, skel_pts[i], line_pts[0])
            right_profile, _ = \
                profile_line(noise, skel_pts[i], line_pts[1])
            noise_profile = np.append(left_profile[::-1], right_profile) * \
                bright_unit
        else:
            noise_profile = None

        if distance is not None:
            total_dists = np.append(-left_dists[::-1], right_dists) \
                * u.pix * phys_per_pix
        else:
            total_dists = np.append(-left_dists[::-1], right_dists) \
                * u.pix * deg_per_pix

        if noise is not None:
            if len(total_profile) != len(noise_profile):
                raise ValueError("Intensity and noise profile lengths do not"
                                 " match. Have you applied the same mask to"
                                 " both?")

        line_profiles.append(total_profile)
        line_distances.append(total_dists)
        profile_extents.append([line_pts[0], skel_pts[i], line_pts[1]])

        if fit_profiles:
            # Now fit!
            profile_fit, profile_fit_err, red_chisq = \
                gauss_fit(total_dists.value, total_profile.value,
                          sigma=noise_profile)

            profile_fits.append(np.hstack([profile_fit, profile_fit_err]))
            red_chisqs.append(red_chisq)

        if verbose:
            p.subplot(121)
            p.imshow(image, origin='lower')
            p.contour(skeleton, colors='r')
            p.plot(skel_pts[i][1], skel_pts[i][0], 'bD')
            p.plot(line_pts[0][1], line_pts[0][0], 'bD')
            p.plot(line_pts[1][1], line_pts[1][0], 'bD')

            p.subplot(122)
            p.plot(total_dists, total_profile, 'bD')
            pts = np.linspace(total_dists.min().value,
                              total_dists.max().value, 100)
            if fit_profiles:
                p.plot(pts, gaussian(pts, *profile_fit), 'r')

            if distance is not None:
                unit = (u.pix * phys_per_pix).unit.to_string()
            else:
                unit = (u.pix * deg_per_pix).unit.to_string()
            p.xlabel("Distance from skeleton (" + unit + ")")
            p.ylabel("Surface Brightness (" + bright_unit.to_string() + ")")
            p.tight_layout()
            p.show()

    if fit_profiles:
        profile_fits = np.asarray(profile_fits)
        red_chisqs = np.asarray(red_chisqs)

        # Create an astropy table of the fit results
        param_names = ["Amplitude", "Std Dev", "Background"]
        param_errs = [par + " Error" for par in param_names]
        colnames = param_names + param_errs
        in_bright_units = [True, False, True] * 2
        tab = QTable()

        tab["Number"] = np.arange(profile_fits.shape[0])
        tab.add_index("Number")

        tab["Red Chisq"] = red_chisqs

        for i, (name, is_bright) in enumerate(zip(colnames, in_bright_units)):
            if is_bright:
                col_unit = bright_unit
            else:
                if distance is not None:
                    col_unit = (u.pix * phys_per_pix).unit
                else:
                    col_unit = (u.pix * deg_per_pix).unit

            tab[name] = profile_fits[:, i] * col_unit

        return line_distances, line_profiles, profile_extents, tab
    else:
        return line_distances, line_profiles


def perpendicular(a):
    '''
    Return the perpendicular vector to a given 2D vector.
    '''
    b = np.empty_like(a)
    b[0] = -a[1]
    b[1] = a[0]
    return b


def walk_through_skeleton(skeleton):
    '''
    Starting from one end, walk through a skeleton in order. Intended for use
    with skeletons that contain no branches.
    '''

    # Calculate the end points
    end_pts = return_ends(skeleton)
    if len(end_pts) != 2:
        raise ValueError("Skeleton must contain no intersections.")

    # Force the first end point to be closest to the image origin.
    if two_point_dist(end_pts[1], [0, 0]) < two_point_dist(end_pts[0], [0, 0]):
        end_pts = end_pts[::-1]

    all_pts = int(np.sum(skeleton))

    yy, xx = np.mgrid[-1:2, -1:2]
    yy = yy.ravel()
    xx = xx.ravel()

    for i in range(all_pts):
        if i == 0:
            ordered_pts = [end_pts[0]]
            prev_pt = end_pts[0]
        else:
            # Check for neighbors
            y, x = prev_pt
            # Extract the connected region
            neighbors = skeleton[y - 1:y + 2, x - 1:x + 2].ravel()
            # Define the corresponding array indices.
            yy_inds = yy + y
            xx_inds = xx + x

            hits = [int(elem) for elem in np.argwhere(neighbors)]
            # Remove the centre point and any points already in the list
            for pos, (y_ind, x_ind) in enumerate(zip(yy_inds, xx_inds)):
                if (y_ind, x_ind) in ordered_pts:
                    hits.remove(pos)

            num_hits = len(hits)

            if num_hits == 0:
                # You've reached the end. It better be the other end point
                if prev_pt[0] != end_pts[1][0] or prev_pt[1] != end_pts[1][1]:
                    raise ValueError("Final point does not match expected"
                                     " end point. Check input skeleton for"
                                     " intersections.")
                break
            elif num_hits == 1:
                # You have found the next point
                posn = hits[0]
                next_pt = (y + yy[posn], x + xx[posn])
                ordered_pts.append(next_pt)
            else:
                # There's at least a couple neighbours (for some reason)
                # Pick the 4-connected component since it is the closest
                for fours in four_conn_posns:
                    if fours in hits:
                        posn = hits[hits.index(fours)]
                        break
                else:
                    raise ValueError("Disconnected eight-connected pixels?")
                next_pt = (y + yy[posn], x + xx[posn])
                ordered_pts.append(next_pt)
            prev_pt = next_pt

    return ordered_pts


def return_ends(skeleton):
    '''
    Find the endpoints of the skeleton.
    '''

    end_points = []

    for i, struct in enumerate(end_structs):
        hits = nd.binary_hit_or_miss(skeleton, structure1=struct)

        if not np.any(hits):
            continue

        for y, x in zip(*np.where(hits)):
            end_points.append((y, x))

    return end_points


def find_path_ends(posn, max_dist, vector):
    '''
    Find ends of a path for line_profile given
    a vector direction.
    '''

    vector = vector.astype(float) / np.linalg.norm(vector)

    max_size = np.ceil(max_dist).astype(int)

    yy, xx = np.mgrid[-max_size:max_size + 1, -max_size:max_size + 1]

    max_circle = yy**2 + xx**2 <= max_dist**2
    ring = \
        np.logical_xor(max_circle,
                       nd.binary_erosion(max_circle, eight_conn))

    radius_pts = [(y + posn[0], x + posn[1]) for y, x in zip(*np.where(ring))]

    x_step = vector[1]
    y_step = vector[0]

    y_diff = max_size * y_step
    x_diff = max_size * x_step

    neg_line_posn = (posn[0] - y_diff, posn[1] - x_diff)
    pos_line_posn = (posn[0] + y_diff, posn[1] + x_diff)

    # pos_dists = np.array([two_point_dist(pos_line_posn, pt)
    #                       for pt in radius_pts])
    # neg_dists = np.array([two_point_dist(neg_line_posn, pt)
    #                       for pt in radius_pts])

    # These should be the ones used to ensure
    # proper distance from the skeleton point.
    # pos_posn give weird results though...

    # pos_posn = radius_pts[np.argmin(pos_dists)]
    # neg_posn = radius_pts[np.argmin(neg_dists)]

    return [neg_line_posn, pos_line_posn]


def two_point_dist(pt1, pt2):
    return np.linalg.norm([x2 - x1 for x2, x1 in zip(pt1, pt2)])


def avg_pts(pts):
    dims = len(pts[0])
    avg_pt = []
    for dim in range(dims):
        avg_pt.append(np.mean([pt[dim] for pt in pts]).astype(int))
    return avg_pt


def gaussian(x, *p):
    '''
    Parameters
    ----------
    x : list or numpy.ndarray
        1D array of values where the model is evaluated
    p : tuple
        Components are:
        * p[0] Amplitude
        * p[1] Mean
        * p[2] Width
        * p[3] Background
    '''
    return (p[0] - p[2]) * np.exp(-1 * np.power(x, 2) /
                                  (2 * np.power(p[1], 2))) + p[2]


def gauss_fit(distance, rad_profile, sigma=None):
    '''
    Fits a Gaussian to the radial profile.

    Parameters
    ----------
    distance : np.ndarray
        Distances from the skeleton.
    rad_profile : np.ndarray
        Intensity values from the image.
    sigma : np.ndarray
        Errors to be used for the fit. Assumes these are 1-sigma errors.

    Returns
    -------
    fit : numpy.ndarray
        Fit values.
    fit_errors : numpy.ndarray
        Fit errors.
    red_chisq : float
        The reduced chi-squared value for the fit.
    '''

    p0 = (np.max(rad_profile), np.std(distance), np.min(rad_profile))

    try:
        fit, cov, info, _, _ = \
            curve_fit(gaussian, distance, rad_profile, p0=p0,
                      maxfev=100 * (len(distance) + 1), sigma=sigma,
                      absolute_sigma=True, full_output=True)
        fit_errors = np.sqrt(np.diag(cov))
        red_chisq = (info['fvec']**2).sum() / (len(info['fvec']) - len(fit))

    except Exception as e:
        print("curve_fit failed with " + str(e))
        # p.plot(distance, rad_profile)
        # p.show()

        fit = np.asarray([np.NaN] * len(p0))
        fit_errors = np.asarray([np.NaN] * len(p0))
        red_chisq = np.NaN

    return fit, fit_errors, red_chisq



import numpy as np
from scipy import ndimage as ndi

'''
Copyright (C) 2011, the scikit-image team
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 3. Neither the name of skimage nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''


def profile_line(img, src, dst, linewidth=1,
                 order=1, mode='constant', cval=0.0):
    """Return the intensity profile of an image measured along a scan line.

    Parameters
    ----------
    img : numeric array, shape (M, N[, C])
        The image, either grayscale (2D array) or multichannel
        (3D array, where the final axis contains the channel
        information).
    src : 2-tuple of numeric scalar (float or int)
        The start point of the scan line.
    dst : 2-tuple of numeric scalar (float or int)
        The end point of the scan line.
    linewidth : int, optional
        Width of the scan, perpendicular to the line
    order : int in {0, 1, 2, 3, 4, 5}, optional
        The order of the spline interpolation to compute image values at
        non-integer coordinates. 0 means nearest-neighbor interpolation.
    mode : {'constant', 'nearest', 'reflect', 'mirror', 'wrap'}, optional
        How to compute any values falling outside of the image.
    cval : float, optional
        If `mode` is 'constant', what constant value to use outside the image.

    Returns
    -------
    return_value : array
        The intensity profile along the scan line. The length of the profile
        is the ceil of the computed length of the scan line.

    Examples
    --------
    >>> x = np.array([[1, 1, 1, 2, 2, 2]])
    >>> img = np.vstack([np.zeros_like(x), x, x, x, np.zeros_like(x)])
    >>> img
    array([[0, 0, 0, 0, 0, 0],
           [1, 1, 1, 2, 2, 2],
           [1, 1, 1, 2, 2, 2],
           [1, 1, 1, 2, 2, 2],
           [0, 0, 0, 0, 0, 0]])
    >>> profile_line(img, (2, 1), (2, 4)) # doctest: +SKIP
    array([ 1.,  1.,  2.,  2.])

    Notes
    -----
    The destination point is included in the profile, in contrast to
    standard numpy indexing.
    """
    perp_lines = _line_profile_coordinates(src, dst, linewidth=linewidth)
    if img.ndim == 3:
        pixels = [ndi.map_coordinates(img[..., i], perp_lines,
                                      order=order, mode=mode, cval=cval)
                  for i in range(img.shape[2])]
        pixels = np.transpose(np.asarray(pixels), (1, 2, 0))
    else:
        pixels = ndi.map_coordinates(img, perp_lines,
                                     order=order, mode=mode, cval=cval)
    # intensities = pixels.mean(axis=1)
    intensities = np.nanmean(pixels, axis=1)

    # Return the distances from the source.
    dists = np.sqrt((perp_lines[0].mean(1) - src[0])**2 +
                    (perp_lines[1].mean(1) - src[1])**2)

    # Check for NaNs and remove those points
    if np.isnan(intensities).any():
        dists = dists[np.isfinite(intensities)]
        intensities = intensities[np.isfinite(intensities)]

    return intensities, dists


def _line_profile_coordinates(src, dst, linewidth=1):
    """Return the coordinates of the profile of an image along a scan line.

    Parameters
    ----------
    src : 2-tuple of numeric scalar (float or int)
        The start point of the scan line.
    dst : 2-tuple of numeric scalar (float or int)
        The end point of the scan line.
    linewidth : int, optional
        Width of the scan, perpendicular to the line

    Returns
    -------
    coords : array, shape (2, N, C), float
        The coordinates of the profile along the scan line. The length of the
        profile is the ceil of the computed length of the scan line.

    Notes
    -----
    This is a utility method meant to be used internally by skimage functions.
    The destination point is included in the profile, in contrast to
    standard numpy indexing.
    """
    src_row, src_col = src = np.asarray(src, dtype=float)
    dst_row, dst_col = dst = np.asarray(dst, dtype=float)
    d_row, d_col = dst - src
    theta = np.arctan2(d_row, d_col)

    length = int(np.ceil(np.hypot(d_row, d_col) + 1))
    # we add one above because we include the last point in the profile
    # (in contrast to standard numpy indexing)
    line_col = np.linspace(src_col, dst_col, length)
    line_row = np.linspace(src_row, dst_row, length)

    # we subtract 1 from linewidth to change from pixel-counting
    # (make this line 3 pixels wide) to point distances (the
    # distance between pixel centers)
    col_width = (linewidth - 1) * np.sin(-theta) / 2
    row_width = (linewidth - 1) * np.cos(theta) / 2
    perp_rows = np.array([np.linspace(row_i - row_width, row_i + row_width,
                                      linewidth) for row_i in line_row])
    perp_cols = np.array([np.linspace(col_i - col_width, col_i + col_width,
                                      linewidth) for col_i in line_col])
    return np.array([perp_rows, perp_cols])
